# -*- coding: utf-8 -*-
"""
Created on Sun Jul 1 15:42:00 2019
@author: ju-bar

Image data access tools

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
# %%
from numba import jit # include compilation support
import numpy as np # include numeric functions
# %%
def image_load(str_file_name, nx, ny, datatype):
    '''
    Loads data from a file and reshapes to an image of ny rows and row length nx.
    '''
    img0 = np.fromfile(str_file_name, dtype=datatype).reshape((ny,nx))
    return img0
# %%
def image_diffractogram(image):
    '''
    Returns the diffractogram I(q)**2 of an image
    '''
    ndim = image.shape
    assert len(ndim)==2, 'this is for 2d images only'
    ndim2 = np.array([ndim[0]>>1,ndim[1]>>1]) # get nyquist index
    imgft = np.fft.fft2(image) # fourier transform
    ftpowsca = 1. / (ndim[0] * ndim[1])
    dif0 = (imgft.real**2 + imgft.imag**2) * ftpowsca # absolute square
    dif1 = np.roll(dif0, shift = ndim2, axis = (0,1)) # put dc on [ny/2,nx/2]
    return dif1
# %%
def image_at_nn(image, pos):
    '''

    Returns nearest-neighbor interpolation of image at position pos.

    '''
    dimg = image.shape
    ic = np.floor(pos[::-1]).astype(int) # reverse sequence of coordinates
    j = np.mod(ic[0],dimg[0])
    i = np.mod(ic[1],dimg[1])
    v = image[j,i]
    return v
# %%
@jit
def image_at_bilin(image, pos):
    '''

    Returns a bi-linear interpolation of image at position pos.

    '''
    dimg = image.shape
    il = int(np.floor(pos[0])) # x
    ih = il + 1
    fi = pos[0] - il
    jl = int(np.floor(pos[1])) # y
    jh = jl + 1
    fj = pos[1] - jl
    # periodic
    il = np.mod(il,dimg[1]) # x
    ih = np.mod(ih,dimg[1])
    jl = np.mod(jl,dimg[0]) # y
    jh = np.mod(jh,dimg[0])
    # sum to bi-linear interpolation
    v = image[jl,il] * (1-fj) * (1-fi) + image[jl,ih] * (1-fj) * fi \
        + image[jh,il] * fj * (1-fi) + image[jh,ih] * fj * fi
    return v
# %%
def image_at(image, pos, ipol=1):
    '''

    Returns the image intensity at a given image position.

    Parameters:
        image : numpy.array of 2 dimensions
            data on a regular grid
        pos : numpy.array of 1 dimension
            2-dimensional coordinate (x,y)
        ipol : int
            interpolation order, default = 1
            0 = nearest neighbor
            1 = bi-linear

    Remark:
        Doesn't check for reasonable input ranges.
        Applies periodic boundary conditions

    '''
    ipol_switcher = {
        0 : image_at_nn(image, pos),
        1 : image_at_bilin(image, pos)
    }
    return ipol_switcher.get(ipol, image_at_bilin(image, pos))
# %%
@jit
def image_resample(image, nout, p0in=np.array([0.,0.]), 
                   p0out=np.array([0.,0.]), 
                   sampling=np.array([[1.,0.],[0.,1.]]), ipol=1):
    '''

    Resamples an image on a grid of new dimensions with given shift and sampling

    Parameters:
        image : numpy.array of 2 dimensions
            data on a regular grid
        nout : array (size=2)
            new dimensions (ny,nx)
        p0in : numpy.array, size=2
            reference position in input array (x,y)
        p0out : numpy.array, size=2
            reference position in output array (x,y)
        sampling : numpy.array, shape(2,2)
            sampling matrix transforming a position in the output to
            a position in the input
        ipol : integer
            interpolation order

    '''
    image_out = np.zeros((nout[0],nout[1]))
    for j in range(0, nout[0]):
        for i in range(0, nout[1]):
            pout = np.array([i,j]) - p0out
            pin = np.dot(sampling,pout) - p0in
            image_out[j,i] = image_at(image, pin, ipol)
    return image_out
# %%
@jit
def image_pos_sum(image, lpoints, ipol=1):
    '''

    Calculates the sum of image intensities at points given by lpoints.

    Parameters:
        image : numpy.array of 2 dimensions
            data on a regular grid
        lpoints : numpy.array of 2 dimensions
            list of 2-dimensional coordinates defining points in image
        ipol : integer
            interpolation order

    Returns:
        Sum of intensities at the given points

    '''
    dpts = lpoints.shape
    npts = dpts[0]
    spt = 0.
    for l in range(0, npts):
        spt += image_at(image, lpoints[l], ipol)
    return spt

def maxpos(array):
    '''
    Returns the index of the first array position with maximum value

    Parameters:
        array : numpy.ndarray int, float
            list of values

    Return:
        numpy.array (2,)
            (x,y) location of global maximum
    '''
    nd = array.shape
    assert len(nd)==2, 'this works only for 2d arrays'
    imax = np.argmax(array)
    i = imax%nd[1]
    j = int((imax-i)/nd[1])
    return np.array([i,j])

def com(array):
    '''
    Returns the center of mass of image data

    Parameters:
        array : numpy.ndarray int, float
            list of values

    Return:
        numpy.array (2,)
            (x,y) center of mass
    '''
    nd = array.shape
    assert len(nd)==2, 'this works only for 2d arrays'
    s0 = 0.
    sx = 0.
    sy = 0.
    for j in range(0, nd[0]):
        for i in range(0, nd[1]):
            v = array[j,i]
            s0 += v
            sx += v * i
            sy += v * j
    return np.array([sx/s0,sy/s0])

# %%
@jit
def data_convolute_2d(sfile_fmt, ix, iy, ndata, krn, thr_krn):
    dkrn = krn.shape
    my = dkrn[0]
    mx = dkrn[1]
    apat = np.zeros(ndata).astype(np.float32) # init zero accumulation buffer
    for jy in range(0, my):
        ly = (jy - iy) % my # periodic wrap of the the shifted y index of the kernel
        for jx in range(0, mx):
            lx = (jx - ix) % mx # periodic wrap of the the shifted x index of the kernel
            if (krn[ly,lx] > thr_krn): # only add effective pixels
                sfile_in = sfile_fmt.format(jx,jy)
                bpat = np.fromfile(sfile_in, dtype=np.float32)
                apat += bpat * krn[ly,lx]
    return apat
# %%
def convolute_2d(data, kernel):
    """

    Convolutes the 2d array data by the 2d kernel.

    Parameters
    ----------
        data : numpy.ndarray
            2-dimensional input data
        kernel : numpy.ndarray
            2-dimensional input kernel

    """
    nd = data.shape
    nk = kernel.shape
    assert len(nd)==2, 'input data must have two dimensions'
    assert len(nk)==2, 'input kernel must have two dimensions'
    assert (nd[0]==nk[0] and nd[1]==nk[1]), 'input data and kernel must be of identical size'
    data_ft = np.fft.fft2(data)
    kernel_ft = np.fft.fft2(kernel)
    return np.fft.ifft2(data_ft * kernel_ft.conjugate())
# %%
def convolute_img_ser(img_ser, kernel):
    nd = img_ser.shape
    nk = kernel.shape
    assert (len(nd)==2 or len(nd)==3), 'input data must have two or three dimensions'
    assert len(nk)==2, 'input kernel must have two dimensions'
    if (len(nd)==2): return convolute_2d(img_ser, kernel).real
    assert (nd[1]==nk[0] and nd[2]==nk[1]), 'input images and kernel must be of identical size'
    kernel_ft = np.fft.fft2(kernel)
    img_out = img_ser.copy()
    for iimg in range(0, nd[0]):
        data_ft = np.fft.fft2(img_ser[iimg])
        img_out[iimg,:,:] = np.fft.ifft2(data_ft * kernel_ft.conjugate())
    return img_out


# -*- coding: utf-8 -*-
"""
Created on Wed Sep 23 08:53:00 2020
@author: ju-bar

Constants and functions related to electrons and their physics

Declaration of physical constants based on published CODATA 2014
Source:	<http://dx.doi.org/10.1103/RevModPhys.88.035009>

Constant symbols are declared with capital letters.
Prefixes:
PHYS_ : general physics constants
EL_ : electron related data

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
import numpy as np
# Exact constants
PHYS_C = 299792458. # vacuum speed of light [ m s^(-1) ]
PHYS_MU0 = 1.2566370614359172953850573533118E-6 # magnetic constant = 4*Pi * 10^(-7) [ N A^(-1) ]
PHYS_EPS0 =	8.8541878176203898505365630317108E-12 # electric constant = 1/(_MU0 _C0^2) [ F m^(-1) ]

# Derived constants
PHYS_HPL = 6.626070040E-34 # Planck's constant, error (81) in last digits [ J s ]
PHYS_HPLEV = 4.135667662E-15 # Planck's constant, error (25) in last digits [ eV s ]
PHYS_HBAR = 1.054571800E-34 # reduced Planck's constant, error (13) in last digits [ J s ]
PHYS_HBAREV = 6.582119514E-16 # reduced Planck's constant, error (40) in last digits [ eV s ]
PHYS_QEL = 1.6021766208E-19 # elementary charge, error (98) in last digits [ C ]

EL_M0 = 9.10938356E-31 # electron rest mass, error (11) in last digits [ kg ]
EL_E0 = EL_M0 * PHYS_C**2 # electron rest energy [ J ]
EL_E0EV = EL_E0 / PHYS_QEL # electron rest energy [ eV ]
EL_E0KEV = EL_E0 / PHYS_QEL / 1000. # electron rest energy [ keV ]

EL_WLKEV = PHYS_C * PHYS_HPL / PHYS_QEL * 1.0E+6 # electron wave length scale from energy [ nm keV ]

EL_CFFA = EL_M0 * PHYS_QEL**2 / (2.0 * PHYS_HPL**2) / (4. * np.pi * PHYS_EPS0) * 1.0E-10 # Coulombic form factor amplitude : m0 e^2 / ( 2 h^2 ) / ( 4 Pi eps0 ) *10^-10 [ -> A^-1 ]

def ht2wl(ht):
    '''
    Calculates the electron wavelength in nm from the kinetic energy (ht) in keV

    Parameters
    ----------
        ht : float
            beam energy in keV

    Return
    ------
        lambda : float
            electron wavelength in nm
    '''
    return EL_WLKEV / np.sqrt(ht * (ht + 2. * EL_E0KEV))

def relcor(ekv):
    '''
    Calculates the relativistic correction factor gamma for electrons of energy ekin [keV]

    Parameters
    ----------
        ht : float
            beam energy in keV

    Return
    ------
        float : rc = (m0*c^2 + ekv) / (m0*c^2)

    '''
    return (EL_E0KEV + ekv) / EL_E0KEV

def k2theta(q, e0, de=0.):
    """

    Returns the scattering angle theta for scattering with momentum transfer hbar*q.
    An energy loss de>0 can be specified to calculate scattering angles for inelastic scattering.

    Parameters
    ----------
        q : float
            length of the vector of momentum transfer devided by hbar in 1/nm units
        e0 : float
            electron kinetic energy in eV
        de : float
            electron energy loss in eV, default: 0

    Return
    ------
        theta : float
            scattering angle in radians [0, pi]
    """
    theta = 0.
    if (np.abs(de) >= e0):
        print("Warning (k2theta): energy not conserved, de is larger or equal e0.")
        return theta
    k0 = 1. / ht2wl(e0 * 0.001) # electron wave number (total momentum)
    kp = 1. / ht2wl((e0-np.abs(de)) * 0.001) # electron wave number after scattering (by conservation of energy)
    if (q < k0-kp):
        print("Warning (k2theta): momentum not conserved, q = {:.5E} is smaller than k0 - k' = {:.5E} ".format(q, k0-kp))
        return theta
    if (q > k0+kp):
        print("Warning (k2theta): momentum not conserved, q = {:.5E} is larger than k0 + k' = {:.5E} ".format(q, k0+kp))
        return np.pi
    # draw the triangle of k0 and kp enclosing theta and q opposite to theta
    # apply the law of cosine
    # q**2 = k0**2 + kp**2 - 2 * k0 * kp * cos(theta)
    # solve for cos(theta)
    costheta = (k0**2 + kp**2 - q**2) / (2 * k0 * kp)
    # invert
    theta = np.arccos(costheta)
    return theta

def kperp2theta(qperp, e0, de=0.):
    """

    Returns the scattering angle theta for a given perpendicular component of the scattering vector.
    Assumes forward scattering.

    Parameters
    ----------
        qperp : float
            perpendicular component of the scattering vector in 1/nm units
        e0 : float
            electron kinetic energy in eV
        de : float
            electron energy loss in eV, default: 0

    Return
    ------
        theta : float
            scattering angle in radians [0, pi]
    """
    theta = 0.
    if (np.abs(de) >= e0):
        print("Warning (kperp2theta): energy not conserved, de is larger or equal e0.")
        return theta
    kp = 1. / ht2wl((e0-np.abs(de)) * 0.001) # electron wave number after scattering (by conservation of energy)
    # vector of kp goes at some angle theta with respect to vector k0,
    # with qperp beeing the transversal component of vector kp.
    # thus sin(theta) = qperp / kp
    theta = np.arcsin(qperp/kp)
    return theta

def theta2qperp(theta, e0, de=0.):
    """

    Returns the perpendicular component of the scattering vector for
    a given scattering angle theta.

    Parameters
    ----------
        theta : float
            scattering angle in radians
        e0 : float
            electron kinetic energy in eV
        de : float
            electron energy loss in eV, default: 0

    Return
    ------
        qperp : float
            perpendicular component of the scattering angle in 1/nm
    """
    qperp = 0.
    assert ((0 <= theta) and (theta <=np.pi)), 'this assumes 0 <= theta <= pi'
    if (np.abs(de) >= e0):
        print("Warning (theta2qperp): energy not conserved, de is larger or equal e0.")
        return qperp
    kp = 1. / ht2wl((e0-np.abs(de)) * 0.001) # electron wave number after scattering (by conservation of energy)
    # vector of kp goes at some angle theta with respect to vector k0,
    # with qperp beeing the transversal component of vector kp.
    # thus sin(theta) = qperp / kp
    qperp = np.sin(theta) * kp
    return qperp


# -*- coding: utf-8 -*-
"""
Created on Tue Sep 22 15:16:00 2020
@author: ju-bar

Functions related to optical focus spread

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
from numba import jit # include compilation support
import numpy as np
from emilys.image.polar import polar_transform
from scipy.optimize import curve_fit
# %%
def dist_gaussian(x, delta):
    '''

    Calculates values of a Gaussian focus spread distribution with
    focus spread 1/e-half-width delta.

    Parameters:
        x : float
            defocus value
        delta : float
            1/e-half-width of the focal distribution

    Returns:
        float

    '''
    return np.exp(-x**2 / delta**2) / (np.sqrt(np.pi) * np.abs(delta))

# %%
def info_limit(delta, lamb):
    '''

    Calculates the information limit for a Gaussian focus-spread with
    1/e half-width delta and for electrons of wavelength lamb.

    Parameters:
        delta : float
            focus spread 1/e-half-width
        lamb : float
            electron wave-length

    Returns:
        float : info-limit (same unit as delta and lamb)

    '''
    return ((np.pi*lamb*delta)**2 / 8.)**0.25
# %%
def delta_of_kappa(kappa, g, t, lamb):
    '''
    Calculate the focus-spread parameter from the circ_lf model
    parameter kappa for a given reciprocal space circle radius g
    beam tilt magnitude t and electron wavelength lamb
    '''
    return np.sqrt(0.5 * kappa) / (np.pi * lamb * g * t)
# %%
def kappa_of_delta(delta, g, t, lamb):
    '''
    Calculate the kappa parameter for the circ_lf model from
    focus spread delta for a given reciprocal space circle radius g
    beam tilt magnitude t and electron wavelength lamb
    '''
    return 2.0 * (np.pi * lamb * g * t * delta)**2

# %%
def circ_lf(phi, phi_t, g, t, c0, c1, kappa):
    '''
    
    Calculates a value of the low-frequency diffractogram component
    under tilted beam illumination in polar notation for points (g, phi) 
    in the diffractogram, beam tilts (t, phi_t), assuming an additional
    base line c0, a modulation amplitude c1 and focus spread dependent
    parameter kappa (see delta_of_kappa).
    The model parameters c0, c1 and kappa are usually not known and
    can be determined from values of the low-frequency content scanned
    along circles around the origin.

    Parameters:
        phi : float
            diffractogram point azimuth in radians
        phi_t : float
            beam tilt azimuth in radians
        g : float
            circle spatial frequency
        t : float
            beam tilt magnitude in units of a spatial frequency
        c0 : float
            base line value
        c1 : float
            contrast amplitude
        kappa : float
            width parameter depending on focus spread
    
    Returns: float

    '''
    return c0 + c1 * np.exp(-kappa * np.cos(phi - phi_t)**2) * np.cosh(kappa * g / t * np.cos(phi - phi_t))
# %%
def circ_mod(xdata, xt, c0, c1, kap):
    '''
    Circular model of the low-frequency components with tilted beam illumintion.
    
    Parameters:
        xdata : numpy array of dimension 2
            xdata[0,:] = x values to calculate for (azimuth)
            xdata[1,0] = g value of the fit
            xdata[2,0] = t value of the fit
        xt : float
            tilt azimuth
        c0 : float
            base line
        c1 : float
            amplitude
        kap : float
            target parameter kap = 2.0 * (np.pi * lamb * g * t * delta)**2
    '''
    x, g, t = xdata[0,:], xdata[1,:], xdata[2,:]
    return circ_lf(x, xt, g[0], t[0], c0, c1, kap)
# %%
def measure_fs_lf(lf_dif, samp_q, q_rng, tilt, lamb):
    '''

    Measure the focus spread assuming a Gaussian focal distribution from
    the low-frequency component of a diffractogram recorded under tilted
    beam illumination with a thin amorphous sample and significant defocus.
    The analysis will be done on rings in the pattern with radius q_min
    to q_max. A square size input lf_dif is assumed with isotropic sampling.

    Parameters:
        lf_dif : numpy.array of 2 dimensions (square size assumed)
            low-frequency component of the diffractogram with Thon-rings
            effectively filtered out
        samp_q : float
            sampling rate of the diffractogram data (e.g. 1/nm / pixel)
            along both grid directions (isotropic sampling assumed)
        q_rng : array of length 2 (q_min, q_max)
            Range of spatial frequencies used in the analysis in the same
            unit as samp_q (e.g. 1/nm). Check that there is no Thon-ring
            modulation left in lf_dif over this range.
        tilt : array of length 2 (tilt_x, tilt_y)
            beam tilt applied for the measurement in the same unit as samp_q
            and q_rng (e.g. 1/nm).
        lamb : float
            electron wavelength in the reciprocal unit of samp_q (e.g. nm).

    Returns:
        delta, err_delta : 2 floats
            focus spread and its error estimate

    Remarks:
        1) designed to primarily work with nm and 1/nm units
        2) input data dimension should be at least 256 x 256

    '''
    nd = lf_dif.shape
    assert len(nd)==2, 'parameter 1 must be a 2d array'
    ndim = nd[0]
    assert (ndim > 32 and nd[1]==ndim), 'parameter 1 must be of some square size (>32)'
    assert samp_q > 0., 'parameter 2 must be larger than zero'
    ndim2 = ndim >> 1
    q_max = ndim2 * samp_q
    assert len(q_rng)==2, 'parameter 3 must have length 2'
    assert (q_rng[1] < q_max and q_rng[0] <= q_rng[1]), 'parameter 3 is not a valid range compared to parameter 1 and 2'
    assert len(tilt)==2, 'parameter 4 must have length 2'
    t_mod = np.sqrt(tilt[0]**2 + tilt[1]**2)
    assert t_mod > 0., 'parameter 4 is a vector of zero length'
    t_phi = np.arctan2(tilt[1],tilt[0])
    assert lamb > 0., 'parameter 5 is not a valid electron wavelength'
    d_q = q_rng[1] - q_rng[0]
    nrad = 1 + (int(d_q / samp_q + 0.5) >> 1) # number of rings from (range / sampling / 2)
    nphi = 180 # 180 samples along the azimuth
    porg = np.array([ndim2,ndim2]).astype(float) # origin of the pattern data in lf_dif
    prng = q_rng / samp_q # radial range in pixels
    arng = np.array([0.,2*np.pi]) # full azimuth range
    apol = polar_transform(lf_dif, nrad, nphi, porg, prng, arng) # get polar transform (could also use polar_resample)
    # --- fit model to all azimuthal curves ---
    xvals = np.zeros((3,nphi)) # prepare x-array
    xvals[0,:] = np.arange(0,nphi) * 2.0 * np.pi / nphi # set the azimuth values as x
    xvals[2,0] = t_mod # tilt modulus (same for all rings)
    qvals = np.arange(0,nrad) * (q_rng[1] - q_rng[0]) / nrad + q_rng[0] # set the radial (q) values
    lprm = np.zeros((nrad, 4)) # prepare array to store fit results for all rings
    lcov = np.zeros((nrad, 4, 4)) # prepare array to store covariance matrix for all rings
    ldel = np.zeros((nrad, 2)) # prepare array storing focus spread and its error for each ring
    for irow in range(0, nrad): # loop over all rings
        yvals = apol[irow] # get values on current ring
        ymin = np.amin(yvals) # minimum
        yamp = np.amax(yvals) - ymin # amplitude
        kap0 = kappa_of_delta(4., qvals[irow], t_mod, lamb) # kappa preset
        xvals[1,0] = qvals[irow] # q of ring
        popt, pcov = curve_fit(circ_mod, xvals, yvals, p0=np.array([t_phi, ymin, yamp, kap0])) # fit from presets
        lprm[irow,:] = popt # store best fitting parameters
        lcov[irow,:,:] = pcov # store covariance matrix
        delta = delta_of_kappa(popt[3], qvals[irow], t_mod, lamb) # calculate focus spread
        ldel[irow,0] = delta # store focus spread
        ldel[irow,1] = 0.25 * delta**2 * pcov[3,3] / popt[3]**2 # variance propagation from kappa
    m_phi = np.mean(lprm[:,0]) # mean tilt orientation
    v_phi = np.mean(lcov[:,0,0]) # tilt orientation variance
    if (np.abs(m_phi - t_phi) > 2.*np.sqrt(v_phi)):
        print('Detected tilt orientation:', m_phi, 'rad')
        print('   input tilt orientation:', t_phi, 'rad')
    m_del = np.mean(ldel[:,0]) # mean delta
    v_del_in = np.mean(ldel[:,1]) # mean variance from fits
    v_del_ex = np.var(ldel[:,0]) # variance over rings
    v_del = max(v_del_in, v_del_ex) # use larger variance
    s_del = np.sqrt(v_del)
    return m_del, s_del, qvals, ldel # return mean and std. deviation for delta and the g-resolved measurement


# -*- coding: utf-8 -*-
"""
Created on Fri Jun 28 08:42:00 2019
@author: ju-bar

Calculations and data handling related to the Modulation Transfer Function (MTF)
pixelated imaging detectors.


This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
# %%
from numba import jit # include compilation support
import numpy as np # include numeric functions
from scipy.optimize import curve_fit
# %%
def loadmtf(sfilename):
    """

    Loads a list of values from a text file and returns a numpy array
    which is a list of tuples (frequency,mtf(frequency)).

    Parameters:
        filename: string
            The name of the file to load MTF data from

    Remarks:
        The file structure is expected to contain a number in the
        first row, specifying, which of the values following below
        corresponds to the Nyquist frequency (0.5) of the detector.
        If the detector has N pixels per row, this should be the
        item N/2 + 1.
        The routine attempts reading values beyond Nyquist, if present.

    """
    file_handle = None
    opened = False
    try:
        file_handle = open(sfilename, 'r') # try open the file
    except (OSError,IOError) as esc:
        print("Error: loadmtf failed:",format(esc))
    else:
        opened = True
        str_lines = file_handle.readlines()
    finally:
        if file_handle: file_handle.close()
        if not opened: return np.array([])

    nlines = len(str_lines)
    if nlines > 0:
        inyq = int(str_lines[0])-1
        inum = 2*inyq
        print("- number of mtf values: ", nlines-1)
        print("- Nyquist index: ", inyq-1)
        ares = np.zeros((nlines-1,2))
        for i in range(1,nlines):
            ares[i-1,0] = float(i-1)/inum
            ares[i-1,1] = float(str_lines[i])
        return ares
    else:
        return np.array([])
# %%
def getmtfkernel(lmtf, ndim, fsca=1.):
    """

    Calculates an MTF kernel array (numpy.array) for a given
    list of radial MTF values, frequency scale and array dimensions

    Parameters:
        lmtf: np.array(shape=(m,2))
            list of m tuples (frequency, mtf value)
        ndim: np.array(2)
            2D image (kernel) dimensions [ny, nx]
        fsca: float (default = 1.)
            frequency scale corresponding to the ratio of
            sampling rates (target / data) or (experiment / simulation)
            The parameter accepts an array of length 2 defining
            different scales for each dimension [fscay, fscax].

    Remarks:
        The MTF values are assumed to be without pixelation coefficients
        and sampled on an equidistant grid. The frequency sampling rate
        used is read from lmtf[1,0].
        MTF data is extrapolated by an exponential decay from the last
        third of values if provided data has only frequencies below 0.75.
        Pixelation coefficients are multiplied corresponding to the
        grid size defined by ndim.

    """
    # define local functions for tail fitting
    def texp(x, a, b, c): # exponential tail function
        return a * np.exp(-b * x) + c
    if ndim.size != 2: return np.array([])
    nmtf = lmtf.shape # number of available mtf data
    if len(nmtf) != 2: return np.array([])
    if nmtf[0] < 10: return np.array([])
    lfsca = np.array([1.,1.])
    if (np.size(fsca) > 1): # different scale for x and y
        lfsca = np.array(fsca.flatten()[0:2])
    else: # same scale for x and y
        lfsca[0] = np.array([fsca]).flatten()[0]
        lfsca[1] = lfsca[0]
    # initialize helper variables and arrays
    dfmtf = lmtf[1,0] # mtf frequency sampling
    fy = np.fft.fftfreq(ndim[0]) # y-frequency indices
    fx = np.fft.fftfreq(ndim[1]) # x-frequency indices
    amtf = np.zeros((ndim[0],ndim[1])) # prepare mtf array
    lmtf1 = lmtf # initialized used mtf data
    # mtf span check
    if lmtf[int(nmtf[0])-1,0] / np.max(lfsca) < 0.75: # do we need to extrapolate beyond the given mtf data ?
        # yes, extrapolate to f = 1.0
        lmtft = lmtf.T # get transposed data
        # fit to the last 1/3 values
        nfitrng = np.array([2*int(nmtf[0]/3),int(nmtf[0])-1]) # set fit index range
        fdata = lmtf[nfitrng[0]:nfitrng[1]].T # get data for fitting
        popt = curve_fit(texp, fdata[0], fdata[1]) # fit
        # extrapolate frequency range to 1 from the fitted exponential
        n1 = 1 + int(np.max(lfsca)/dfmtf + 0.5) # determine index range for frequencies up to 1
        # print('- extrapolation to f =', np.max(lfsca), ', index =', n1)
        fextr = np.arange(nfitrng[1]+1,n1) * dfmtf # prepare frequency array
        mtfextr = texp(fextr, *popt[0]) # calculate tail values
        lmtf1 = np.array([np.append(lmtft[0],fextr),np.append(lmtft[1],mtfextr)]).T # append and transpose back
        nmtf = lmtf1.shape # update number of available mtf data
    # mtf grid calculation
    for j in range(0,ndim[0]): # loop rows (v)
        fy2 = (lfsca[0] * fy[j])**2 # squared row frequency
        scy = np.sinc(fy[j]) # row sinc ! Sinc is for the given sampling, not affected by lfsca
        for i in range(0,ndim[1]): # loop columns (h)
            fm2 = (lfsca[1] * fx[i])**2 + fy2 # frequency modulus squared
            scx = np.sinc(fx[i]) # column sinc
            fm = np.sqrt(fm2) # scaled frequency modulus
            # prepare linear interpolation on radial mtf list
            km = fm/dfmtf # mtf array index corresponding to fm
            k0 = int(km) # lower array index
            fk = km - k0 # fraction to lower array index
            k1 = min(k0 + 1,nmtf[0]) # limited upper array index
            amtf[j,i] = scy * scx * ( (1. - fk) * lmtf1[k0,1] + fk * lmtf1[k1,1] ) # linear interpolation and sinc product
    return amtf


# -*- coding: utf-8 -*-
"""
Created on Thu Dec 10 18:19:00 2020
@author: ju-bar

Functions related to work with crystal structures

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
import numpy as np
from emilys.optics.econst import ht2wl, k2theta, theta2qperp
# %%
def HOLZ_theta(e0, c, n=1, de=0., verbose=False):
    """

    Calculates the scattering angle of the nth-order Laue zone
    for a crystal with lattice constant c along the incident beam
    direction and for scattering with energy loss de.

    Parameters
    ----------

        e0 : float
            electron beam kinetic energy in eV
        c : float
            lattice constant along the beam direction in nm
        n : int
            order of the Laue zone, default: 1
        de : float  
            electron energy loss in eV, default: 0
        verbose : boolean
            flags extra output on (True) or off (False, default)

    Return
    ------
        theta : float
            scattering angle around which to expect HOLZ

    """
    assert n > 0, "this is not working for n < 1."
    theta = 0.
    if (np.abs(de) >= e0):
        print("Warning: energy not conserved, de is larger or equal e0.")
        return theta
    k0 = 1. / ht2wl(e0 * 0.001) # electron wave number (total momentum)
    if (verbose): print('k0 = {:.5E}'.format(k0))
    kp = 1. / ht2wl((e0-np.abs(de)) * 0.001) # electron wave number after scattering (by conservation of energy)
    if (verbose): print('kp = {:.5E}'.format(kp))
    kc = n / c # longitudinal reciprocal lattice vector length (from k0 to the Laue zone)
    if (verbose): print('kc = {:.5E}'.format(kc))
    kz = k0 - kc # residual vertical wave number (from center to Laue zone)
    if (verbose): print('kz = {:.5E}'.format(kz))
    if (kz > kp): # ... k' must be able to reach the Laue zone (forward scattering)
        print("Warning: momentum not conserved, kz is larger than k'")
        return theta
    if (kz < -kp): # ... k' must be able to reach the Laue zone (backward scattering)
        print("Warning: momentum not conserved, kz is smaller than -k'")
        return theta
    kt = np.sqrt(kp**2 - kz**2) # transversal component
    if (verbose): print('kt = {:.5E}'.format(kt))
    q = np.sqrt((kp-kz)**2 + kt**2) # scattering vector on the Laue circle
    if (verbose): print('q  = {:.5E}'.format(q))
    theta = k2theta(q, e0, de)
    return theta

def HOLZ_qperp(e0, c, n=1, de=0., verbose=False):
    """

    Calculates the perpendicular component of the scattering vector
    to the nth-order Laue zone for a crystal with lattice constant c
    along the incident beam direction and for scattering with energy
    loss de.

    Parameters
    ----------

        e0 : float
            electron beam kinetic energy in eV
        c : float
            lattice constant along the beam direction in nm
        n : int
            order of the Laue zone, default: 1
        de : float  
            electron energy loss in eV, default: 0
        verbose : boolean
            flags extra output on (True) or off (False, default)

    Return
    ------
        qperp : float
            perpendicular component of the scattering vector [1/nm]

    """
    return theta2qperp(HOLZ_theta(e0, c, n, de, verbose), e0, de)


# -*- coding: utf-8 -*-
"""
Created on Fri Jul 12 13:03:00 2019
@author: ju-bar

Image shift utilities

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
import numpy as np
from emilys.image.imagedata import maxpos, com
# %%
def get_rigid_shift(img, ref, method='pixel', maxshift=0.):
    '''
    Measures the rigid shift of data in the 2d array 'img' relative to 'ref'.

    Parameters:
        img : numpy.ndarray (M,N)
            test image
        ref : numpy.ndarray (M,N)
            reference image
        method : string, default: 'pixel'
            shift finding method
                'pixel' : closest pixel shift
                'com' : center of mass
                'itcom' : iterative center of mass
                'peak' : peak fitting
                'itpeak' : iterative peak fitting
        maxshift : float, default: 0. = no limit
            limitation of shift vector in pixels along both dimensions

    Return:
        numpy.ndarray (2,)
            rigid shift vector (x,y)
    '''
    lims = False # default correlation maximum search region is not limited
    ls = np.abs(maxshift) # check for search limit
    if (ls > 0.): lims = True # there is a search limit
    assert img.shape == ref.shape and len(img.shape) == 2, \
        'expecting 2d numpy arrays of equal shape as parameters 1 and 2'
    ndim = np.array(img.shape) # get image dimension
    csca = 1. / ndim[0] / ndim[1] # get rescaling for fft
    mi = np.mean(img) # get image mean
    si = np.std(img) # get image standard deviation
    mf = np.mean(ref) # get reference mean
    sf = np.std(ref) # get reference standard deviation
    fti = np.fft.fft2((img - mi)/si) # FT image
    ftr = np.fft.fft2((ref - mf)/sf) # FT reference
    iorg = (ndim / 2).astype(int) # get origin
    cfir0 = np.roll(np.real(np.fft.ifft2(fti * np.conjugate(ftr))), iorg, np.array([0,1])) * csca # cross-correlation rolled to origin
    dc = iorg
    if lims: # extract part of the x-corr image to look for maximum
        ils = np.ceil(ls)
        i00 = int(max(0, iorg[0]-ils))
        i01 = int(min(ndim[0], iorg[0]+ils))
        i10 = int(max(0, iorg[1]-ils))
        i11 = int(min(ndim[1], iorg[1]+ils))
        dc = iorg - np.array([i00,i10]) # move center
        cfir = cfir0[i00:i01+1,i10:i11+1].copy() # copy the data
    else:
        cfir = cfir0.copy() # copy all data
    method_switch = { # decide how to determine the maximum location
        'pixel' : maxpos(cfir) - np.flip(dc),
        'com' : com(cfir) - np.flip(dc)
    }
    return method_switch.get(method, maxpos(cfir) - np.flip(dc))
    


#%%



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 02 17:02:00 2020
@author: ju-bar

Routines for powder diffraction data analysis.

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
# %%
from numba import jit # include compilation support
import numpy as np
from emilys.image.arrayplot import arrayplot2d
from matplotlib.patches import Circle
# %%
def get_DSrings_radi(ndim, hkl, a0, dpix, mag):
    '''

    Calculate the radius in pixels of Debye-Scherrer rings in a diffractogram of an image
    of a powder sample recorded with a given magnification. Assumes a square size image
    and a cubic lattice.

    Parameters:
        ndim : int
            size of the image in pixels
        hkl : array of shape (n,3) of type int
            list of [h,k,l] indices of a cubic reciprocal lattice
        a0 : float
            real lattice constant
        dpix : float
            image pixel size (same unit as a0)
        mag : float
            magnification factor of the image

    Returns:
        array (n) of floats
            list of radii of Debye-Scherrer rings in the 2D Fourier transform of the image.

    '''
    nring = len(hkl)
    lq = np.zeros(nring)
    lr = lq.copy()
    for i in range(0, nring): # generate list of diffraction vectors related to the lattice
        lq[i] = np.sqrt(np.dot(hkl[i],hkl[i])) / a0
        lr[i] = lq[i] * dpix * ndim / mag
    return lr
# %%
def plot_DSrings_on_dif(dif, hkl, a0, dpix, mag, psca = 0.5, pamp = 0.25, prng = [0.,0.5]):
    '''

    Plots a diffractogram (FT amplitudes of an image) and overlays rings corresponding to
    the reciprocal lattice vectors given by the list hkl for a cubic crystal lattice of
    lattice constant a0. An image pixel size dpix and magnification used to record the image
    must be given. pamp is an exponent used to display the intensities and prng is the
    relative range of intensities used for display scaling colors from min to max.

    '''
    ndim = dif.shape[0]
    ndim2 = ndim >> 1
    nring = len(hkl)
    lr = get_DSrings_radi(ndim, hkl, a0, dpix, mag)
    pdif = arrayplot2d(dif**pamp, psca, 'inferno', vrange = prng)
    porg = [ndim2, ndim2]
    print(lr)
    for i in range(0, nring):
        if (lr[i] < ndim2):
            circ = Circle(porg, lr[i], color='b', linestyle=(5, (10, 4)), fill=False)
            pdif[1].add_patch(circ)


# -*- coding: utf-8 -*-
"""
Created on Sun Jul 1 17:15:00 2019
@author: ju-bar

Cluster recognition tools

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
import numpy as np
# %%
class cluster:
    '''
    
    class cluster

    Handles the assignment of a list item to a cluster, 
    tracking mean and standard deviation of cluster members.

    Members:
        mean : array_like
            mean value of cluster items
        sdev : array_like
            standard deviation of cluster items
        population : array
            list of initial data indices populating the cluster
    
    '''
    def __init__(self, mean, sdev, population):
        self.mean = mean
        self.sdev = sdev
        self.population = population

    def update_stats(self, ldata):
        val0 = ldata[0] * 0.
        self.mean = val0
        self.sdev = val0
        tmpsum = val0
        tmpsqrs = val0
        npop = len(self.population)
        if npop > 0:
            for i in self.population:
                tmpval = ldata[i]
                tmpsum = tmpsum + tmpval
                tmpsqrs = tmpsqrs + tmpval**2
            self.mean = tmpsum / npop
            if npop > 1:
                self.sdev = np.sqrt(np.abs(tmpsqrs / npop - self.mean**2)) \
                            * npop / (npop - 1)

    def add_member_index(self, ldata, index):
        self.population.append(index)
        self.update_stats(ldata)

    def merge(self, other, ldata):
        self.population.extend(other.population)
        self.update_stats(ldata)

class cluster_assignment:
    '''

    Class cluster_assignment

    Handles the cluster assignment and is returned from cluster
    finding algortithms.

    Members:
        lassign : array, int
            forward assignment list (data index -> cluster index)
        lcluster : array, cluster
            list of cluster objects (cluster -> data indices)

    '''
    def __init__(self, nitems):
        '''
        Initialize a cluster_assignment object for a given number of data items.
        '''
        self.lassign = np.full(nitems, -1) # initialize forward list with no assignment
        self.lcluster = [] # initialize empty cluster list

    def add_to_cluster(self, idx_cluster, idx_item, ldata):
        '''
        Adds a data item to an existing cluster.
        '''
        ncl = len(self.lcluster)
        nitems = len(self.lassign)
        if (idx_cluster >= 0 and idx_cluster < ncl and idx_item >= 0 \
            and idx_item < nitems):
            if self.lassign[idx_item] < 0:
                self.lcluster[idx_cluster].add_member_index(ldata, idx_item)
                self.lassign[idx_item] = idx_cluster
                return 0 # success
            return 20 # error, double assignment
        return 10 # error, invalid index

    def add_cluster(self, cluster):
        '''
        Add a cluster, if no conflict is found. Returns list of conflicting items.
        '''
        lconfl = [] # init list of assignment conflicts
        newpopnum = len(cluster.population)
        if newpopnum > 0:
            for i in cluster.population:
                if self.lassign[i] >= 0: # this item already has a cluster
                    lconfl.append(i)
            if len(lconfl) == 0: # no conflict -> add cluster
                self.lcluster.append(cluster)
                icl = len(self.lcluster) - 1
                for i in cluster.population:
                    self.lassign[i] = icl # set forward assignment
                return [] # return empty list of conflicts
            else: # ! conflict, return the list of conflicts
                return lconfl
        return [] # return empty list of conflicts

    def del_cluster(self, idx):
        '''Removes a cluster.'''
        ncl = len(self.lcluster)
        if idx >= 0 and idx < ncl:
            self.lcluster.pop(idx)
            # update the assignment list
            for i in range(0, len(self.lassign)):
                if self.lassign[i] == idx:
                    self.lassign[i] = -1 # erase assignment
                if self.lassign[i] > idx:
                    self.lassign[i] -= 1 # decrement assignment to clusters following the deleted

    def merge_clusters(self, itarget, isource):
        '''
        Merges cluster isource to cluster itarget and updates the assignments.
        
        Returns list of conflicting items.
        '''
        ncl = len(self.lcluster)
        lconfl = []
        if (itarget >=0 and isource >= 0 and itarget < ncl and isource < ncl):
            cl_target = self.lcluster[itarget]
            cl_source = self.lcluster[isource]
            cl_merge = cl_target
            cl_merge.merge(cl_source)
            self.del_cluster(itarget)
            self.del_cluster(isource)
            lconfl = self.add_cluster(cl_merge)
            if len(lconfl) > 0: # This should never happen
                # try reversing the deletion
                self.add_cluster(cl_target)
                self.add_cluster(cl_source)
        return lconfl
# %%
def cluster_l2(ldata, l2_thresh=1.):
    '''

    Creates a list of clusters found in ldata with the l2-norm
    of cluster members better than l2_thresh with respect
    to the cluster mean.

    Parameters:
        ldata : numpy.ndarray, shape(N,2)
            list of 2d point coordinates
        l2_thresh: float
            clustering threshold, must be > 0!

    Return:

        clusters are returned as list of values
            [mean,sdev,population]
            mean : array_like as items of ldata
                mean value of the cluster
            sdev : array_like as items of ldata
                standard deviation of the cluster
            population : array
                list of indices referencing to items of ldata in the cluster
        
        An empty list is returned in case of invalid parameters.

    '''
    nd = ldata.shape
    if len(nd) != 2:
        return []
    ndat = nd[0]
    ndim = nd[1]
    lthr = np.abs(l2_thresh)
    if lthr == 0.:
        return []
    lass = cluster_assignment(ndat)
    nass = 0 # initialize number of assigned values
    for i in range(0, ndat): # loop over all items
        iclclose = -1 # remember index of closest cluster
        dclclose = 1000. * lthr # initialize closest cluster distance to very large
        if len(lass.lcluster) > 0: # try to find possible clusters
            for j in range(0, len(lass.lcluster)): # loop over current clusters
                vdist = lass.lcluster[j].mean - ldata[i] # cluster mean distance vector to query point
                dist = np.sqrt(np.dot(vdist,vdist)) # cluster mean distance
                if dist < l2_thresh: # cluster mean distance is below threshold
                    if dist < dclclose: # closest cluster ?
                        dclclose = dist # remember
                        iclclose = j
        if iclclose < 0: # no good cluster found
            # create and add a new cluster
            cl_new = cluster(ldata[i],np.zeros(ndim),[i])
            if len(lass.add_cluster(cl_new)) == 0: 
                nass += 1 # increase number of assigned items
        else: # cluster found
            if 0 == lass.add_to_cluster(iclclose, i, ldata): # populate the cluster
                nass += 1 # increase number of assigned items
    print('- ',nass,'of',ndat,' items assigned to ', \
          len(lass.lcluster),' clusters')
    return lass



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 19 10:55:00 2020
@author: ju-bar

Image data manipulation routines

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
# %%
from numba import jit # include compilation support
import numpy as np # include numeric functions
# %%
def replace_area_from_random_surrounding(image, pos, r1, r2):
    '''

    Replaces image data at point pos within radius r1 from randomly
    chosen data point in range r1 to r2.
    A modified copy of image is returned, image is not changed.

    Parameters:
        image : numpy.array of 2 dimensions
            data on a regular grid
        pos : numpy.array of 1 dimension
            2-dimensional coordinate (x,y)
        r1 : float
            radius of replacement destinations around pos
        r2 : float
            radius of replacement sources around pos
    '''
    ndim = image.shape
    assert len(ndim)==2, 'this works only with a 2d image'
    assert len(pos)==2, 'pos must specify a 2d position'
    assert (pos[0] >= 0 and pos[0] <= ndim[1]-1 and pos[1] >= 0 and pos[1] <= ndim[0]-1), 'pos must be inside the image'
    assert r2 > 1., 'r2 must be larger than 1'
    assert r1 + 1 < r2, 'r2 must be at least 1 pixel larger than r1'
    assert (r2 < ndim[0]/2 and r2 < ndim[1]/2), 'r2 must be smaller than image half size'
    # init
    image_out = image.copy()
    # determine bounding box of replacements to avoid looping through the whole image
    rpbx0 = max(0,int(np.floor(pos[0]-r1)))
    rpbx1 = min(ndim[1]-1,int(np.ceil(pos[0]+r1)))
    rpby0 = max(0,int(np.floor(pos[1]-r1)))
    rpby1 = min(ndim[0]-1,int(np.ceil(pos[1]+r1)))
    # loop over bounding box and replace all pixels with distances smaller or equal r1 from pos
    for j in range(rpby0, rpby1+1):
        dy = j - pos[1]
        dy2 = dy * dy
        for i in range(rpbx0, rpbx1+1):
            dx = i - pos[0]
            r = np.sqrt(dx * dx + dy2)
            if (r > r1): continue # outside r1, skip pixel
            # current pixel value will be replaced
            # dice random replacement until one is found in the image
            while True:
                srnd = np.random.rand(2)
                sprad = srnd[0] * (r2 - r1) + r1
                spphi = srnd[1] * 2. * np.pi
                spi = int(np.round(pos[0] + sprad * np.cos(spphi)))
                spj = int(np.round(pos[1] + sprad * np.sin(spphi)))
                if (spi >= 0 and spi < ndim[1] and spj >= 0 and spj < ndim[0]): break # replacement source found
            # do the replacement
            image_out[j,i] = image[spj,spi]
    # done, return modified copy
    return image_out



# -*- coding: utf-8 -*-
"""
Created on Mon Sep 21 15:14:00 2020
@author: ju-bar

Linear search function.

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
# %%
from numba import jit # include compilation support
import numpy as np
# %%
def linsrch_minloc1(data, subpix = True):
    '''

    Searches for the first local minimum of the input data and returns the index
    where that occurs. A 2nd-order sub-pixel estimate is attempted if
    subpix == True and the initial search didn't stop at either end of the data row.

    Parameters:
        data : numpy.array
            input data, rows of data will be searched, 2d input is supported
        subpix : bool
            switches sub-pixel estimates

    Returns: float or 1d numpy.array depending on dimension of input data

    '''
    ndim = data.shape
    assert (len(ndim)>0 and len(ndim)<3), 'this only supports 1d or 2d input data'
    nrows = 1
    if len(ndim)==2:
        b2d = True
        nrows = ndim[0]
        ndat = ndim[1]
    else:
        b2d = False
        ndat = ndim[0]
    dout = np.zeros(nrows, dtype=float)
    for j in range(0, nrows):
        if (b2d): # set drow from 2d input
            drow = data[j]
        else: # set drow from 1d input
            drow = data
        # find 1st local minimum in drow
        im = 0
        vmin = drow[0]
        for i in range(1, ndat): # loop row data
            if (drow[i] < vmin): # new min
                vmin = drow[i]
                im = i
            else:
                break # data increases, stop
        dout[j] = 1.0 * im # store min loc
        if (subpix and im > 0 and im < ndat-1): # try to estimate min-loc sub-sample
            # determine minimum location of the parabole crossing points (im-1, im, im+1)
            v1 = drow[im-1]
            v2 = drow[im]
            v3 = drow[im+1]
            dout[j] = (4. * im**2 + v1 - v3 +  2. * im * (v1 - 2. * v2 + v3))/(2. * (v1 - 2. * v2 + im * (2. + v1 - v3) + v3))
    #
    if (nrows == 1): return dout[0] # return scalar
    return dout # return vector

# %%
def maxloc_com(y, yerr=0., x0=-1., xrng = 10.):
    calc_err = False
    nd = y.shape
    assert (len(nd)<=2 and len(nd)>=1), 'expecting 1d or 2d array is input data'
    if (np.size(yerr)==np.size(y)):
        nds = yerr.shape
        assert len(nds)==len(nd), 'input y and yerr must be of same dimension'
        assert nds[0]==nd[0], 'input y and yerr must be of same shape'
        if (len(nd)==2):
            assert nds[1]==nd[1], 'input y and yerr must be of same shape'
        calc_err = True
    nrow = 1
    if (len(nd) == 2):
        nrow = nd[0]
        ndim = nd[1]
    else:
        ndim = nd[0]
    ly = y.reshape((nrow,ndim))
    lys = yerr.reshape((nrow,ndim))
    if (np.size(x0) == nrow):
        lx0 = x0.copy()
    else:
        lx0 = np.argmax(ly,axis=1)
    # ---- #
    lc = np.zeros((nrow,2)).astype(np.float64)
    lc[:,0] = lx0
    lit = np.zeros(nrow).astype(int)
    lrep = np.zeros(nrow).astype(int) + 1
    nrep = np.sum(lrep) # get number of rows to keep on iterating
    while (nrep > 0):
        lcprev = lc.copy() # previous result
        for irow in range(0, nrow):
            if (lrep[irow] == 0): continue # skip row, it is not marked for further iteration
            lit[irow] += 1
            lxi = np.arange(0,ndim).astype(np.float64)
            lwi = np.exp( - (lxi-lc[irow,0])**2 / (2. * xrng**2))
            lyi = ly[irow]
            m0 = np.sum(lwi*lyi)
            m1 = np.sum(lwi*lyi*lxi)
            lc[irow,0] = m1 / m0
            if (calc_err): # error propagation
                lsi = lys[irow]
                s1 = np.sum((lsi*lwi*lxi)**2) 
                s2 = np.sum((lsi*lwi)**2) * m1**2 / m0**2
                s3 = -2. * np.sum((lsi*lwi)**2 * lxi) * m1 / m0
                lc[irow,1] = (s1 + s2 + s3) / m0**2
            if ((np.abs(lc[irow,0] - lcprev[irow,0]) < 0.001) or lit[irow] > 100): # convergence check
                lrep[irow] = 0 # set halt to iteration for this row
        # ----- loop on rows
        nrep = np.sum(lrep) # get number of rows to keep on iterating
    # --- iterator for convergence
    print(lit)
    return lc


# -*- coding: utf-8 -*-
"""
Created on Wed Sep 23 10:21:00 2020
@author: ju-bar

Routines for Thon ring analysis.

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
# %%
from numba import jit # include compilation support
import numpy as np
from emilys.image.polar import polar_resample
from emilys.numerics.linsearch import linsrch_minloc1
from emilys.image.kernels import filter_normal_iso_2d, filter_normal_2d, filter_exp4_2d
from scipy.optimize import curve_fit
# %%
@jit
def lf_component(diffractogram):
    '''
    Calculates the low-frequency diffractogram component by adaptive filtering of the
    auto-correlation function. This should return an image, where Thon rings modulations
    are effectively filtered out, except for the first ring.

    Parameters:
        diffractogram : numpy.array of dimension 2
            input diffractogram data

    Result:
        numpy.array of the same size as diffractogram

    Remarks:
        1) The low-frequency component in the auto-correlation function is assumed to be smaller than 10% Nyquist.
        2) An anisotropic filter will be used adopting to the shape of the inner peak of the auto-correlation function.

    '''
    ndim = diffractogram.shape
    assert (len(ndim)==2), 'this is for 2d input arrays only'
    ndim2 = np.array([ndim[0]>>1,ndim[1]>>1]) # get nyquist index
    nxy = 0.5 * (ndim2[0] + ndim2[1])
    adifft = np.fft.ifft2(diffractogram)
    acf0 = np.abs(adifft)  # amplitudes of the ft
    acf1 = np.roll(acf0, shift = ndim2, axis = (0,1)) # recenter the acf
    # extract the polar transform of the central part
    crng = 1 + int(0.1 * nxy) # central range limiting low-frequency search
    apolc = polar_resample(acf1,crng,32,ndim2,np.array((0.,1.*crng)),np.array((0.,2*np.pi))).T
    # determine the extent of the central peak in the auto-correlation function
    lwcpk = linsrch_minloc1(apolc,False)
    # print(lwcpk)
    # calculate a low-pass filter kernel for extracting the diffractogram low-frequency part
    flt_rad_max = np.amax(lwcpk)
    flt_rad_min = np.amin(lwcpk)
    flt_dir = np.angle(np.fft.fft(lwcpk)[2]) + 0.5 * np.pi
    alpflt0 = filter_normal_2d(ndim,[1.,1.],flt_rad_max,flt_rad_min,flt_dir)
    # apply low-pass filtering
    return np.real(np.fft.fft2(adifft * alpflt0))
# %%
@jit
def lp_filter(diffractogram, freq):
    '''
    calculates a radial low-pass filter of the input with frequency
    limit freq in units of nyquist using a Gaussian filter function

    diffractogram : numpy.array of 2 dimensions
        input image / diffractogram
    freq : relative filter strength (0 ... 1)
    '''
    ndim = diffractogram.shape
    assert (len(ndim)==2), 'this is for 2d input arrays only'
    ndim2 = np.array([ndim[0]>>1,ndim[1]>>1]) # get nyquist index
    nxy = 0.5 * (ndim2[0] + ndim2[1])
    aft = np.fft.fft2(diffractogram)
    flt_rad = freq * nxy
    aff = filter_normal_iso_2d(ndim,[1.,1.],flt_rad)
    return np.fft.ifft2(aff * aft).real



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 17 17:30:00 2020
@author: ju-bar

Polar image transformation

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
# %%
from numba import jit # include compilation support
import numpy as np # include numeric functions
import emilys.image.imagedata as aimg # include image access routines
# %%
#@jit
def polar_resample(image, num_rad, num_phi, pole, rng_rad, rng_phi = [0.,2. * np.pi], ipol = 1):
    '''

    Transforms a 2D image to new grid in polar representation using interpolation.

    Parameters:
        image : numpy.array of 2 dimensions
            data on a regular grid
        num_rad : integer
            number of radial sample
        num_phi : integer
            number of azimuthal samples
        pole : numpy.array, size=2
            position (x,y) of the pole in the input image (fractional pixel coordinate)
        rng_rad : numpy.array, size=2
            radial range in pixels
        rng_phi : numpy.array, size=2
            azimuthal range in radian, use only values from the positive interval [0, 2*Pi]!
            default = np.array((0.,2.*np.pi))
        ipol : int
            interpolation type
            0 = nearest neigbor
            1 = bi-linear

    Returns:
        numpy.array of 2 dimensions, shape (num_rad,num_phi)

    Remarks:
        May produce strange signal with noise data and of the polar sampling is fine
        compared to the input grid.
        This assumes isotropic sampling of the input.

    '''
    nd = image.shape
    # check input
    assert len(nd)==2, 'this works only for 2d arrays as input'
    assert num_rad>=1, 'number of radial samples must be at least 1'
    assert num_phi>=1, 'number of azimuthal samples must be at least 1'
    assert len(rng_rad)==2, 'invalid length of radial range input'
    assert len(rng_phi)==2, 'invalid length of azimuthal range input'
    assert (rng_rad[0]<rng_rad[1] and rng_rad[0]>=0.), 'invalid radial range parameter'
    assert (ipol>=0 and ipol<=1), 'invalid interpolation switch (0,1)'
    # initialize
    ny = nd[0]
    nx = nd[1]
    r0 = rng_rad[0]
    r1 = rng_rad[1]
    p0 = rng_phi[0]
    p1 = rng_phi[1]
    if (p0 == p1): p1 = p0 + 2.*np.pi # same phi -> full circle
    image_out = np.zeros((num_rad,num_phi)) # polar data
    # loop over polar coordinates
    for ir in range(0, num_rad): # loop over radii
        r = r0 + (r1 - r0) * ir / num_rad # radial coordinate
        for ip in range(0, num_phi): # loop over azimuth
            p = p0 + (p1 - p0) * ip / num_phi# azimuthal coordinate
            pos = pole + r * np.array([np.cos(p), np.sin(p)])
            if ((pos[0] >= 0) and (pos[0] <= nx-1) and (pos[1] >= 0) and (pos[1] <= ny-1)):
                image_out[ir,ip] = aimg.image_at(image, pos, ipol)
    return image_out
# %%
@jit
def polar_transform(image, num_rad, num_phi, pole, rng_rad, rng_phi = np.array((0.,2. * np.pi))):
    '''

    Transforms a 2D image to new grid in polar representation using a re-binning algorithm.

    Parameters:
        image : numpy.array of 2 dimensions
            data on a regular grid
        num_rad : integer
            number of radial sample
        num_phi : integer
            number of azimuthal samples
        pole : numpy.array, size=2
            position (x,y) of the pole in the input image (fractional pixel coordinate)
        rng_rad : numpy.array, size=2
            radial range in pixels
        rng_phi : numpy.array, size=2
            azimuthal range in radian, use only values from the positive interval [0, 2*Pi]!
            default = np.array((0.,2.*np.pi))

    Returns:
        numpy.array of 2 dimensions, shape (num_rad,num_phi)

    Remarks:
        Due to the re-binning algorithm, some bins may not receive data, especially when
        using fine sampling around the pole region. If you want to avoid this, you may use
        the routine polar_resample.
    '''
    nd = image.shape
    # check input
    assert len(nd)==2, 'this works only for 2d arrays as input'
    assert num_rad>=1, 'number of radial samples must be at least 1'
    assert num_phi>=1, 'number of azimuthal samples must be at least 1'
    assert len(rng_rad)==2, 'invalid length of radial range input'
    assert len(rng_phi)==2, 'invalid length of azimuthal range input'
    assert (rng_rad[0]<rng_rad[1] and rng_rad[0]>=0.), 'invalid radial range parameter'
    # initialize
    tpi = 2.*np.pi # 2*Pi
    ny = nd[0]
    nx = nd[1]
    r0 = rng_rad[0]
    r1 = rng_rad[1]
    p0 = rng_phi[0]
    p1 = rng_phi[1]
    #norm_out = np.zeros((num_rad,num_phi)) # polar bin population
    image_out = np.zeros((num_rad,num_phi)) # polar data
    check_phi = 0
    delta_p = p1 - p0 # input azimuth range
    if (np.abs(delta_p)>=tpi or np.abs(delta_p)==0.): # full circle
        delta_p = tpi
    else: # not a full circle
        check_phi = 1
    # determine bounding box around the annular segment added to the polar transform
    bx0 = np.min(np.array((r0 * np.cos(p0), r0 * np.cos(p1), r1 * np.cos(p0), r1 * np.cos(p1))) + pole[0]) # left-most corner of segment
    bx1 = np.max(np.array((r0 * np.cos(p0), r0 * np.cos(p1), r1 * np.cos(p0), r1 * np.cos(p1))) + pole[0]) # right-most corner of segment
    by0 = np.min(np.array((r0 * np.sin(p0), r0 * np.sin(p1), r1 * np.sin(p0), r1 * np.sin(p1))) + pole[1]) # lowest corner of segment
    by1 = np.max(np.array((r0 * np.sin(p0), r0 * np.sin(p1), r1 * np.sin(p0), r1 * np.sin(p1))) + pole[1]) # highest corner of segment
    # print("- b-box: ((", bx0, ",", by0, "),(", bx1, ",", by1, "))")
    if (p0 <= 0. and p1 >= 0.): bx1 = pole[0] + r1 # include x = x0 + r1
    if (p0 <= np.pi and p1 >= np.pi): bx0 = pole[0] - r1 # include x = x0 - r1
    if (p0 <= 0.5*np.pi and p1 >= 0.5*np.pi): by1 = pole[1] + r1 # include y = y0 + r1
    if (p0 <= 3.*np.pi/2. and p1 >= 3.*np.pi/2.): by0 = pole[1] - r1 # include y = y0 - r1
    # print("- b-box: ((", bx0, ",", by0, "),(", bx1, ",", by1, "))")
    ix0 = np.min([nx-1, np.max([0, int(np.floor(bx0))])]) # pixel range clipped to image
    ix1 = np.min([nx-1, np.max([0, int( np.ceil(bx1))])])
    iy0 = np.min([ny-1, np.max([0, int(np.floor(by0))])])
    iy1 = np.min([ny-1, np.max([0, int( np.ceil(by1))])])
    # print("- bounding box: ((", ix0, ",", iy0, "),(", ix1, ",", iy1, "))")
    # accumulate polar samples to polar bins
    for j in range(iy0, iy1+1):
        dy = 1.*(j - pole[1])
        dy2 = dy*dy
        for i in range(ix0, ix1+1):
            dx = 1.*(i - pole[0])
            ir2 = dx*dx + dy2
            ir = np.sqrt(ir2)
            if (ir < r0 or ir > r1): continue # outside radial range, skip pixel
            ip = p0
            if (ir > 0.): # 
                ip = np.arctan2(dy,dx) % tpi # pixel azimuth -> into range [0, 2 pi]
                ipr = (ip - p0) / delta_p # azimuth relative to rng_phi - 0 ... 1
                if (check_phi): # segment
                    if (ipr < 0. or ipr > 1.): continue # outside azimuthal range, skip pixel
            jr = int(np.round((ir - r0) / (r1 - r0) * num_rad)) # round to next radial bin
            jp = int(np.round(ipr * num_phi)) # round to next azimuthal bin
            #print("i =",i,", j =",j,", r =",ir,", p =", ip,", jr =",jr,", jp =",jp)
            if (jr >= 0 and jr < num_rad and jp >= 0 and jp < num_phi):
                #norm_out[jr,jp] += 1.
                image_out[jr,jp] += image[j,i]
    ## renormalize the output, removed (2020-10-13 JB)
    #for j in range(0, num_rad):
    #    for i in range(0, num_phi):
    #        if (norm_out[j,i]>0.): image_out[j,i] /= norm_out[j,i] # divide by number of accumulated samples to the polar bin
    return image_out
# # %%
# @jit
# def polar_transform_projector(nx, ny, ax, ay, nr, np, p0, rng_r, rng_p):
#     '''
    
#     Calculates a projection matrix from a recangular non-isotropic space grid to a polar grid.

#     Parameters:
#         nx : int
#             number of grid points along x (fast dimension)
#         ny : int
#             number of grid points along y (slow dimension)
#         ax : float
#             total grid size along x (physical unit)
#         ay : float
#             total grid size along y (physical unit)
#         nr : int
#             number of radial output samples (fast dimension)
#         np : int
#             number of azimuthal output samples (slow dimension)
#         p0 : [float,float]
#             position of the pole in the source array (x,y) (physical unit)
#         rng_r : [float,float]
#             radial range of the output (physical unit)
#         rng_p : [float,float]
#             azimuthal range of the output (radians)

#     Returns:
#         An array of dimension (ny, nx, 2) which assigns a tuple [ir, ip] of the
#         polar output to each pixel [iy, ix] of the cartesian input. 
    
#     Remarks:
#         The output projector p is used in the following way with in input a of
#         shape (ny, nx) and an output b of shape (nr, np):
#         for j in range(0,ny):
#             for i in range(0,nx):
#                 ir = p[j,i,0]
#                 ip = p[j,i,1]
#                 if (ir >= 0 and ip >= 0):
#                     b[ir,ip] += a[j,i]

#     '''
#     p = np.zeros((ny,nx,2), dytpe=np.int32) # projector matrix init (iy,ix) -> (ir,ip)
#     p -= 1 # init with -1
#     sy = ay / ny
#     sx = ax / nx
#     sr = (rng_r[1] - rng_r[0]) / nr
#     sp = (rng_p[1] - rng_p[0]) / np
#     for j in range(0, ny):
#         dy = sy * j - p0[1] # y distance to pole
#         dy2 = dy**2
#         for i in range(0, nx):
#             dx = sx * i - p0[0] # x distance to pole
#             dx2 = dx**2
#             r = np.sqrt(dx2 + dy2)
#             p = np.arctan(dy, dx)
#             ir = np.round((r - r0) / sr)
#             ip = np.round()
#     return p
# %%
@jit
def polar_radpol3_transform(image, num_rad, num_phi, pole, rng_rad, rng_phi = np.array((0.,2. * np.pi)), x0 = 0., x1 = 1., x2 = 0., x3 = 0.):
    '''

    Resamples a 2D image to new grid in polar representation with the radial direction
    sampling on a grid with x = x0 + x1 * r + x2 * r**2 + x3 + r**3, where r is the distance
    to the pole in the input image and x the coordinate of the output image along axis = 0.
    The resampled x axis will correspond to a finer step size towards outer regions of the
    input image causing the respective data to be distributed more sparsely in the output.

    Parameters:
        image : numpy.array of 2 dimensions
            data on a regular grid
        num_rad : integer
            number of radial sample
        num_phi : integer
            number of azimuthal samples
        pole : numpy.array, size=2
            position (x,y) of the pole in the input image (fractional pixel coordinate)
        rng_rad : numpy.array, size=2
            radial range in pixels
        rng_phi : numpy.array, size=2
            azimuthal range in radian, use only values from the positive interval [0, 2*Pi]!
            default = np.array((0.,2.*np.pi))
        x0 : float
            radial sampling offset, default = 0.
        x1 : float
            radial sampling linear coefficient, default = 1.
        x2 : float
            radial sampling 2nd order coefficient, default = 0.
        x3 : float
            radial sampling 3rd order coefficient, default = 0.

    '''
    nd = image.shape
    # check input
    assert len(nd)==2, 'this works only for 2d arrays as input'
    assert num_rad>=1, 'number of radial samples must be at least 1'
    assert num_phi>=1, 'number of azimuthal samples must be at least 1'
    assert len(rng_rad)==2, 'invalid length of radial range input'
    assert len(rng_phi)==2, 'invalid length of azimuthal range input'
    assert (rng_rad[0]<rng_rad[1] and rng_rad[0]>=0.), 'invalid radial range parameter'
    # initialize
    tpi = 2.*np.pi # 2*Pi
    ny = nd[0]
    nx = nd[1]
    r0 = rng_rad[0]
    r1 = rng_rad[1]
    xl0 = x0 + x1 * r0 + x2 * r0**2 + x3 * r0**3 # lower radial limit
    xl1 = x0 + x1 * r1 + x2 * r1**2 + x3 * r1**3 # upper radial limit
    p0 = rng_phi[0]
    p1 = rng_phi[1]
    # norm_out = np.zeros((num_rad,num_phi)) # polar bin population
    image_out = np.zeros((num_rad,num_phi)) # polar data
    check_phi = 0
    delta_p = p1 - p0 # input azimuth range
    if (np.abs(delta_p)>=tpi or np.abs(delta_p)==0.): # full circle
        delta_p = tpi
    else: # not a full circle
        check_phi = 1
    # determine bounding box around the annular segment added to the polar transform
    bx0 = np.min(np.array((r0 * np.cos(p0), r0 * np.cos(p1), r1 * np.cos(p0), r1 * np.cos(p1))) + pole[0]) # left-most corner of segment
    bx1 = np.max(np.array((r0 * np.cos(p0), r0 * np.cos(p1), r1 * np.cos(p0), r1 * np.cos(p1))) + pole[0]) # right-most corner of segment
    by0 = np.min(np.array((r0 * np.sin(p0), r0 * np.sin(p1), r1 * np.sin(p0), r1 * np.sin(p1))) + pole[1]) # lowest corner of segment
    by1 = np.max(np.array((r0 * np.sin(p0), r0 * np.sin(p1), r1 * np.sin(p0), r1 * np.sin(p1))) + pole[1]) # highest corner of segment
    # print("- b-box: ((", bx0, ",", by0, "),(", bx1, ",", by1, "))")
    if (p0 <= 0. and p1 >= 0.): bx1 = pole[0] + r1 # include x = x0 + r1
    if (p0 <= np.pi and p1 >= np.pi): bx0 = pole[0] - r1 # include x = x0 - r1
    if (p0 <= 0.5*np.pi and p1 >= 0.5*np.pi): by1 = pole[1] + r1 # include y = y0 + r1
    if (p0 <= 3.*np.pi/2. and p1 >= 3.*np.pi/2.): by0 = pole[1] - r1 # include y = y0 - r1
    # print("- b-box: ((", bx0, ",", by0, "),(", bx1, ",", by1, "))")
    ix0 = np.min([nx-1, np.max([0, int(np.floor(bx0))])]) # pixel range clipped to image
    ix1 = np.min([nx-1, np.max([0, int( np.ceil(bx1))])])
    iy0 = np.min([ny-1, np.max([0, int(np.floor(by0))])])
    iy1 = np.min([ny-1, np.max([0, int( np.ceil(by1))])])
    # print("- bounding box: ((", ix0, ",", iy0, "),(", ix1, ",", iy1, "))")
    # accumulate polar samples to polar bins
    for j in range(iy0, iy1+1):
        dy = 1.*(j - pole[1])
        dy2 = dy*dy
        for i in range(ix0, ix1+1):
            dx = 1.*(i - pole[0])
            ir2 = dx*dx + dy2
            ir = np.sqrt(ir2)
            if (ir < r0 or ir > r1): continue # outside radial range, skip pixel
            ip = p0
            if (ir > 0.): # no angle at r = 0
                ip = np.arctan2(dy,dx) % tpi # pixel azimuth -> into range [0, 2 pi]
                ipr = (ip - p0) / delta_p # azimuth relative to rng_phi - 0 ... 1
                if (check_phi): # segment
                    if (ipr < 0. or ipr > 1.): continue # outside azimuthal range, skip pixel
            x = x0 + x1 * ir + x2 * ir**2 + x3 * ir**3 # corresponding output grid coordinate
            jr = int(np.round((x - xl0) / (xl1 - xl0) * num_rad)) # round to next radial bin
            jp = int(np.round(ipr * num_phi)) # round to next azimuthal bin
            #print("i =",i,", j =",j,", r =",ir,", p =", ip,", jr =",jr,", jp =",jp)
            if (jr >= 0 and jr < num_rad and jp >= 0 and jp < num_phi):
                # norm_out[jr,jp] += 1.
                image_out[jr,jp] += image[j,i]
    # renormalize the output # removed (2020-10-13 JB)
    #for j in range(0, num_rad):
    #    for i in range(0, num_phi):
    #        if (norm_out[j,i]>0.): image_out[j,i] /= norm_out[j,i] # divide by number of accumulated samples to the polar bin
    return image_out
# %%
def radial_bin_mask(ndim, step, pix_pole, nr, radial_range=None):
    """

    Creates a mask linking image pixels to radial bins.

    Parameters
    ----------
        ndim : array of length 2 and type int
            dimensions of input images (num_rows, num_columns)
        step : array of length 2 and type float
            step sizes of input image (step_rows, step_columns)
        pix_pole : array of length 2 and type float
            pixel position of the pole from measuring radius (row, column)
        nr : int
            number of radial samples
        radial_range : array of lenth 2 and type float
            radial range, if None a default [0, nr] is used

    Returns
    -------
        array dimension ndim and type int containing indices of
        the radial 1d array linked to each input image

    """
    #
    msk = np.zeros(ndim, dtype=int)
    #
    if radial_range is None:
        r0 = 0.
        r1 = 1.0 * nr
    else:
        r0 = radial_range[0]
        r1 = radial_range[1]
    #
    for i_y in range(0, ndim[0]):
        y = step[0] * (i_y - pix_pole[0])
        for i_x in range(0, ndim[1]):
            x = step[1] * (i_x - pix_pole[1])
            r = np.sqrt(x**2 + y**2)
            i_r = round((r - r0) / (r1 - r0) * nr)
            if (i_r >= 0) and (i_r < nr):
                msk[i_y,i_x] = i_r
            else:
                msk[i_y,i_x] = -1
    #
    return msk



# -*- coding: utf-8 -*-
"""
Created on Sun Jul 07 17:48:00 2019
@author: ju-bar

Linear least square algorithms.

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""

import numpy as np

class LeastSquareSolution:
    '''
    class LeastSquareSolution

    Data class for communicating the solution x of a linear
    system of equations
        b = A . x
    with more equations, m = length of b
    than unknowns, n = length of x
    Multiple vectors b will cause the same amount of solutions x

    Members:
        x : numpy.ndarray shape(l,n)
            solution vectors
        cov : numpy.ndarray shape(n,n)
            covariance matrix
        chisq : numpy.ndarray shape(l,)
            sum of squared residuals for each solution in x
        m : int
            number of equations
        n : int
            number of parameters
        l : int
            number of solutions
        solver : string
            solving algorithm
        tol : float
            relative singular value strength accepted
        err : string
            error message
        errcode : int
            error code
        
    '''
    def __init__(self):
        self.x = np.array([])
        self.cov = np.array([])
        self.chisq = np.array([])
        self.m = 0
        self.n = 0
        self.l = 0
        self.solver = 'none'
        self.tol = 1.E-8
        self.err = ''
        self.errcode = 0

def linear_lstsq(A, b, tol=1.E-8):
    '''
    Solves the over-determined system of linear equations
        b = A . x
    for x. For the solution we assume pre-weighted equations.
    If weighting is to be applied, please do this outside of
    this routine.

    Parameters:
        A : array_like (M,N), float
            linear system matrix for M equations and N unknowns
        B : array_like (M,) or (L,M), float
            right side vectors
        tol : float
            tolerance for accepting weak singular values relative
            to the strength of the strongest singular value

    Return:
        LeastSquareSolution object
            solution as x member, multiple for L > 1
            covariance matrix as cov
    '''
    sol = LeastSquareSolution()
    mata = np.array(A)
    shpa = mata.shape
    if (len(shpa) != 2):
        sol.errcode = 1
        sol.err = ('input "A" is expected as 2-dimensional array,'+
                   '{:d} dimension(s) found'.format(len(shpa)))
        print('Error in linear_lstsq: ' + sol.err)
        return sol
    sol.m = shpa[0]
    sol.n = shpa[1]
    vb = np.array(b)
    shpb = vb.shape
    if (len(shpb) < 1 or len(shpb) > 2):
        sol.errcode = 2
        sol.err = ('input "b" is expected as 1- or 2-dimensional array,'+
                   '{:d} dimension(s) found'.format(len(shpb)))
        print('Error in linear_lstsq: ' + sol.err)
        return sol
    if (len(shpb) == 2):
        sol.l = shpb[0]
        if (sol.m != shpb[1]):
            sol.errcode = 3
            sol.err = ('dimension 1 of input "b" ({:d}) should be equal'
                       ' to dimension 0 of input A ({:d})'.format(
                        shpb[1],shpa[0]))
            print('Error in linear_lstsq: ' + sol.err)
            return sol
    else: # len(shpb) == 1
        sol.l = 1
        if (sol.m != shpb[0]):
            sol.errcode = 4
            sol.err = ('dimension 0 of input "b" ({:d}) should be equal'
                       ' to dimension 0 of input A ({:d})'.format(
                        shpb[1],shpa[0]))
            print('Error in linear_lstsq: ' + sol.err)
            return sol
    vb = vb.reshape(sol.l, sol.m) # assure shape of solution vectors
    sol.x = np.zeros((sol.l,sol.n)) # prepare solution vectors
    sol.cov = np.zeros((sol.n,sol.n)) # prepare covariance array
    #
    # - implement solver switch here
    sol.solver = 'svd' # using singular value decomposition from numpy.linalg
    matu, diags, matvh = np.linalg.svd(mata, full_matrices=False)
    #print('- U:',matu.shape,', W:',diags.shape,', Vh:',matvh.shape)
    matvt = np.transpose(matvh)
    diagstol = diags
    # edit singular values below tol
    smax = np.max(diags) # largest singular value
    stol = sol.tol * smax # singular value threshold
    for i in range(0, sol.n):
        if (diagstol[i] < stol): # singular value below threshold?
            diagstol[i] = 0. #
    #
    # calculate covariance matrix
    wdiag = np.zeros(sol.n)
    w2diag = np.zeros(sol.n)
    for i in range(0, sol.n): # prepare weigths from singular values
        if (diagstol[i] > 0.):
            wdiag[i] = 1. / diagstol[i]
            w2diag[i] = 1. / (diagstol[i]**2)
    for i in range(0, sol.n):
        for j in range(0, sol.n):
            csum = np.dot(matvt[i] * matvt[j], w2diag)
            # csum = 0.
            # for k in range(0, sol.n):
            #     # csum = csum + matvh[k,i] * matvh[k,j] * w2diag[k]
            #     csum = csum + matvt[i,k] * matvt[j,k] * w2diag[k]
            sol.cov[i,j] = csum
            sol.cov[j,i] = csum
    #
    # calculate solutions and residuals
    btmp = np.zeros(sol.m, dtype=float)
    for k in range(0, sol.l): # ... for all right-side vectors in b
        # solutions by back-substitution
        # for j in range(0, sol.m):
        #     bsum = 0.
        #     if (diagstol[j] > 0.):
        #         # for i in range(0, sol.n):
        #         #     bsum = bsum + matu[j,i] * vb[k,i]
        #         # bsum = bsum * wdiag[j]
        #         bsum = np.dot(matu[j], vb[k]) * wdiag[j]
        #     btmp[j] = bsum
        btmp = np.dot(matu.T, vb[k]) * wdiag # U_T . b
        # for j in range(0, sol.m):
        #     bsum = 0.
        #     for i in range(0, sol.m):
        #         # bsum = bsum + matvh[i,j] * btmp[i]
        #         bsum = bsum + matvt[j,i] * btmp[i]
        #     sol.x[k,j] = bsum
        sol.x[k] = np.dot(matvt, btmp)
        # residuals for the solution by back-projection
        btmp = np.dot(mata, sol.x[k])
        sol.chisq = np.zeros(sol.l)
        sol.chisq[k] = np.sum((btmp - vb[k])**2)
    #
    return sol


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 04 21:43:00 2019
@author: ju-bar

Functions and classes handling the projection from a 2D object plane
into a 2D image plane including distortions.

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
from numba import jit # include compilation support
import numpy as np
from scipy.special import comb
from scipy.optimize import minimize, root
from emilys.numerics.lstsq import linear_lstsq as llstsq

class projection_func_2d:
    '''
    
    class projection_func_2d

    Handles the calculation of a projection transforming a tuple
    (x0,x1) into a projected tuple (y0,y1).

    projection_func objects support arbitrary polynomial orders of
    distortions. The lowest order (0) is a rigid shift, the first
    order corresponds to a 2x2 matrix projection. Initializing
    with more orders than needed will waste computation power.

    The projection is parameterized by a list of coefficients
    [{alk0},{alk1}] defining the projection according to the
    polynomial series

    [y0,y1] = [a000,a001] + [[a100,a010],[a101,a011]].[x0,x1]
        + [[a200,a110,a020],[a201,a111,a021]].[x0**2,2*x0*x1,x1**2]
        + [[a300,a210,a120,a030],[a301,a211,a121,a031]].[x0**3,3*x0**2*x1,3*x0*x1**2,x1**3]
        + ...
        = sum_l sum_k binomial(l+k,l) * x0**l * x1**k * [alk0 , alk1]

    The linear coefficients are listed in lcoeff in the sequence
    [...,[a100,a101],[a010,a011],...]. A linear distortion matrix
    corresponding to [[a100,a010],[a101,a011]] must therefore be
    entered in transposed form.

    '''
    #
    def __init__(self, max_order):
        # maximum polynomial order of the projection function per dimension
        self.__max_order = max(1, int(max_order)) # at least the 1st order is present
        # number of polynomial terms = number of coefficients per dimension
        self.__num_terms = int((self.max_order + 1) * (self.max_order + 2) / 2 )
        # internal coefficient list
        self.__lcoeff = np.zeros((self.num_terms, 2), dtype=float) # preset to 0
        # coefficient / term use flags (functions will return zeros for not flagged terms)
        self.__luse_term = np.full(self.num_terms, 1, dtype=int) # preset to 1
        # coefficient to index list (l,k) -> idx, -1 = no coefficient (not all are used), READ BUT DO NOT WRITE!
        self.__lcoeff_to_idx = np.full((self.max_order + 1,self.max_order + 1), -1, dtype=int)
        # index to coefficient list idx -> (l,k), READ BUT DO NOT WRITE!
        self.__lidx_to_coeff = np.zeros((self.num_terms, 2), dtype=int)
        # binomial factor list idx -> binomial(l+k,l), READ BUT DO NOT WRITE!
        self.__lbinom = np.zeros(self.num_terms, dtype=int)
        # fill the static lists
        idx = 0
        for n in range(0, self.max_order + 1):
            for l in range(n,-1,-1):
                self.__lcoeff_to_idx[n-l,l] = idx
                self.__lidx_to_coeff[idx,0] = l
                self.__lidx_to_coeff[idx,1] = n - l
                self.__lbinom[idx] = comb(n,l,exact=True)
                idx += 1

    def reinitialize(self, max_order=1):
        '''
        Re-initializes to object to the given projection order (min 1 = linear)

        Parameters:
            max_order : int
                order of the projection polynomial
        '''
        self.__init__(max_order)

    @property
    def max_order(self):
        '''
        Maximum order of the polynomial.
        '''
        return self.__max_order

    @property
    def num_terms(self):
        '''
        Number of terms.
        '''
        return self.__num_terms

    def get_lk(self, idx):
        '''
        Returns the polynomial powers (l,k) for the term of given index.
        '''
        if idx >=0 and idx < self.num_terms:
            return self.__lidx_to_coeff[idx]
        return np.array([])

    @property
    def lcoeff(self):
        '''
        Returns a copy of the internal coefficient list.
        '''
        return self.__lcoeff.copy()

    @lcoeff.setter
    def lcoeff(self, lc):
        '''
        Set internal coefficient list.

        Parameters:
            lcoeff : array_like, float, shape = (self.num_terms,2) or shorter
                polynomial coefficients

        '''
        lc1 = np.atleast_1d(lc)
        assert lc1.dtype in [float,int], 'lcoeff needs numbers!'
        nc1 = np.size(lc1)
        if nc1 > 1:
            nc2 = int((nc1 - nc1%2) / 2)
            lc2 = lc1.flatten()
            self.__lcoeff[0:nc2] = lc2[0:2*nc2].reshape(nc2, 2)

    def set_coeff_idx(self, idx, alk0 = None, alk1 = None):
        '''
        Sets coefficients of the 2 dimensions at index idx
        of the current coefficient list

        Parameters:
            idx : int
                list index
            alk0, alk1 : float
                polynomial term coefficients
        '''
        if idx >= 0 and idx < self.num_terms:
            if alk0 != None:
                self.__lcoeff[idx,0] = alk0
            if alk1 != None:
                self.__lcoeff[idx,1] = alk1

    def set_coeff_lk(self, l, k, alk0 = None, alk1 = None):
        '''
        Sets coefficients of the 2 dimensions for the term x0**l x1**k 

        Parameters:
            l, k : int
                polynomial term exponents
            alk0, alk1 : float
                polynomial term coefficients
        '''
        if l >= 0 and k >= 0 and l+k <= self.max_order:
            idx = self.__lcoeff_to_idx[k,l]
            self.set_coeff_idx(idx, alk0, alk1)

    @property
    def luse(self):
        '''
        Returns a copy of the internal usage flag list.
        '''
        return self.__luse_term.copy()

    @luse.setter
    def luse(self, lu):
        '''
        Set internal usage flag list.

        Parameters:
            lu : array_like, int, shape = (self.num_terms,) or shorter
                polynomial term flags

        '''
        lu1 = np.atleast_1d(lu)
        assert lu1.dtype in [float,int], 'luse needs numbers!'
        nu = np.size(lu1)
        if nu > 0:
            lu2 = lu1.flatten().astype(int)
            self.__luse_term[0:nu] = lu2[0:nu]

    def set_use_idx(self, idx, use=1):
        '''
        Sets useage flag for index idx of the current coefficient list (1 by default)

        Parameters:
            idx : int
                list index
            use : int, default: 1
                polynomial term flag
        '''
        if idx >= 0 and idx < self.num_terms:
            self.__luse_term[idx] = int(use)

    def set_use_lk(self, l, k, use=1):
        '''
        Sets useage flag for the term x0**l x1**k (1 by default)

        Parameters:
            l, k : int
                polynomial term exponents
            use : int
                polynomial term flag
        '''
        if l >= 0 and k >= 0 and l+k <= self.max_order:
            idx = self.__lcoeff_to_idx[k,l]
            self.set_use_idx(idx, use)

    def project(self, x, lcoeff = np.array([]), luse=np.array([])):
        '''
        Projects x with a 2d polynomial.

        Parameters:
            x : array_like, float, tuple, shape = (N,2)
                position in object plane [...,[xi0,xi1],...]
            lcoeff : array_like, float, optional
                sequence of projection coefficient interpreted as tuples {[alk0,alk1]}
            luse : array_like, int, optional
                polynomial term flags

        Return:
            numpy.ndarray
                [...,[yi0,yi1],...]
                list of projected positions as array of shape(N, 2) with
                N = number of tuples in parameter x = numpy.size(x)/2

        Remarks:
            You may set the coefficients before calling the project method
            using projection_func_2d.lcoeff = lcoeff.

            The function calculates for each tuple [xi0,xi1]:
            [yi0,yi1] = sum_j=1,M [alk0, alk1] * tlk(xi0,xi1),
            with tlk(xi0,xi1) = binomial(l+k,l) * x0**l * x1**k
            and an internal listing j :-> l,k for each j in range(0, M).
            M = projection_func_2d.num_terms

        '''
        if np.size(lcoeff) > 0: self.lcoeff = lcoeff # set new coefficients if present
        if np.size(luse) > 0: self.luse = luse # set flags
        nx1 = np.size(x)
        assert 0 == nx1%2 and nx1 > 0, 'project expects an even number of coordinates'
        nx = int(nx1 / 2)
        lx = np.reshape(x,(nx,2))
        ly = np.zeros((nx,2), dtype=float) # prepare result arrays
        n = self.max_order
        m = self.num_terms
        fpx = np.full((n+1, 2), 1., dtype=float)
        fpt0 = np.zeros(m, dtype=float)
        for i in range(0, nx): # for all positions
            pwx = lx[i]
            for l in range(1, n+1): # calculate x0**l and x1**l for l = 1 .. n
                fpx[l] = pwx
                pwx = pwx * lx[i]
            fpt = fpt0 # reset terms
            for j in range(0, m): # calculate binomial(l+k,l) x0**l * x1**k for all terms
                if 0 == self.__luse_term[j]: continue # skip term j
                l, k = self.__lidx_to_coeff[j]
                fpt[j] = fpx[l,0] * fpx[k,1] * self.__lbinom[j]
            ly[i] = np.dot(fpt, self.__lcoeff) # project coefficient lists on binomial term list
        return ly

    def project_deriv_lcoeff(self, x, luse=np.array([])):
        '''
        2d projection function derivatives with respect to all projection
        parameters (lcoeff). Since the projection is a linear function of
        the parameters, the result depends on x only.

        Parameters:
            x : array_like, float, tuple, shape = (N,2)
                position in object plane [...,[xi0,xi1],...]
            luse : array_like, int, optional
                polynomial term flags

        Return:
            numpy.ndarray
                [...,[...,[[dyi0/dalk0, dyi1/dalk0],[dyi0/dalk1, dyi1/dalk1]],...],...]
                list of derivatives as numpy.ndarray of shape(N, M, 2, 2) with
                N = number of tuples in parameter x
                M = number of projection terms tlk(xi0,xi1) = binomial(l+k,l) * x0**l * x1**k

        Remarks:
            The output of this method may be used as Jacobian for least-squares methods.
            Since
                dyi0/dalk0 = dyi1/dalk1 = tlk(xi0,xi1), and
                dyi1/dalk0 = dyi0/dalk1 = 0,
            the 2 x 2 items of the output are [[1,0],[0,1]] * tlk(xi0,xi1).
            This redundancy produces the correct shape of the derivative array for the shape
            (M,2) of the parameter list lcoeff. It consumes 4 times more memory than really needed,
            but this should be acceptable since M is usually not very large (<100).
        '''
        if np.size(luse) > 0: self.luse = luse # set flags
        nx1 = np.size(x)
        assert 0 == nx1%2 and nx1 > 0, 'project expects an even number of coordinates'
        nx = int(nx1 / 2)
        lx = np.reshape(x,(nx,2))
        n = self.max_order
        m = self.num_terms
        ldy = np.zeros((nx,m,2,2), dtype=float) # prepare result arrays
        fpt0 = np.zeros(m, dtype=float)
        fpx = np.full((n+1, 2), 1., dtype=float)
        mid = np.array([[[1.],[0.]],[[0.],[1.]]])
        for i in range(0, nx): # for all positions
            pwx = lx[i]
            for l in range(1, n+1): # calculate x0**l and x1**l for l = 1 .. n
                fpx[l] = pwx
                pwx = pwx * lx[i]
            fpt = fpt0 # reset terms
            for j in range(0, m): # calculate tlk(x) = binomial(l+k,l) x0**l * x1**k for all terms
                if 0 == self.__luse_term[j]: continue # skip term j
                l, k = self.__lidx_to_coeff[j]
                fpt[j] = fpx[l,0] * fpx[k,1] * self.__lbinom[j]
            ldy[i] = np.dot(mid,[fpt]).T # writes [...,[[tlk(x),0],[0,tlk(x)]],...]
        return ldy

    def project_deriv_x(self, x, lcoeff=np.array([]), luse=np.array([])):
        '''
        2d projection function derivatives with respect to positions x,
        evaluated for all x.

        Parameters:
            x : array_like, float, tuple, shape = (N,2)
                position in object plane [...,[xi0,xi1],...]
            lcoeff : array_like, float, optional
                sequence of projection coefficient interpreted as tuples {[alk0,alk1]}
            luse : array_like, int, optional
                polynomial term flags

        Return:
            numpy.ndarray
                [...,[[dyi0/dx0,dyi1/dx0],[dyi0/dx1,dyi1/dx1]],...]
                list of derivatives as array of shape(N, 2, 2) with
                N = number of tuples in parameter x
        
        Remarks:
            You may set the coefficients before calling the project_deriv_x method
            using projection_func_2d.lcoeff = lcoeff.

            The output of this method may be used as Jacobian for least-squares methods.
        
        '''
        if np.size(lcoeff) > 0: self.lcoeff = lcoeff # set new coefficients if present
        if np.size(luse) > 0: self.luse = luse
        nx1 = np.size(x)
        assert 0 == nx1%2 and nx1 > 0, 'project expects an even number of coordinates'
        nx = int(nx1 / 2)
        lx = np.reshape(x,(nx,2))
        n = self.max_order
        m = self.num_terms
        ldy = np.zeros((nx,2,2), dtype=float) # prepare result arrays
        fpx = np.full((n+1, 2), 1., dtype=float)
        fptz = np.zeros((m,2), dtype=float) # prepare derivative terms
        for i in range(0, nx): # for all positions
            pwx = lx[i]
            for l in range(1, n+1): # calculate x0**l and x1**l for l = 1 .. n
                fpx[l] = pwx
                pwx = pwx * lx[i]
            fpt = fptz.copy() # reset terms
            for j in range(0, m): # calculate binomial(l+k,l) [l * x0**(l-1) * x1**k, k * x0**l * x1**(k-1)] for all terms
                if 0 == self.__luse_term[j]: continue # skip term j
                l, k = self.__lidx_to_coeff[j]
                fpt0 = 0.
                if l>0: # catch derivative of constant (no element of fpx[l-1] for l-1 = -1)
                    fpt0 = l * fpx[l-1,0] * fpx[k,1] * self.__lbinom[j] # (dtj/dx0)
                fpt1 = 0.
                if k>0: # catch derivative of constant (no element of fpx[k-1] for k-1 = -1)
                    fpt1 = k * fpx[l,0] * fpx[k-1,1] * self.__lbinom[j] # (dtj/dx1)
                fpt[j] = np.array([fpt0,fpt1]) # set derivative tuple of term j
            # project coefficient lists on binomial term derivative list
            ldy[i] = np.dot(fpt.T, self.__lcoeff) # (2,m).(m,2) -> (2,2): [[dyi0/dx0,dyi1/dx0],[dyi0/dx1,dyi1/dx1]]
        return ldy

    def __bp_project(self, x):
        '''
        Internal projection function for the backprojection root finder.

        Call this only from internal method backproject!

        Assumes tuple input shape (2,).
        '''
        return self.project(x)[0]

    def __bp_project_deriv_x(self, x):
        '''
        Internal projection derivative function for the backprojection root finder.

        Call this only from internal method backproject!

        Assumes tuple input shape (2,).
        '''
        return self.project_deriv_x(x)[0]

    def backproject(self, y, lcoeff = np.array([]), luse=np.array([]), tol=1.E-8):
        '''
        Back-projects from y to x for the 2d polynomial p(x) = y.

        Parameters:
            y : array_like, float, tuple, shape = (N,2)
                position in image plane [...,[yi0,yi1],...]
            lcoeff : array_like, float, optional
                sequence of projection coefficient interpreted as tuples {[alk0,alk1]}
            luse : array_like, int, optional
                polynomial term flags

        Return:
            numpy.ndarray
                [...,[xi0,xi1],...]
                list of back-projected positions as array of shape(N, 2) with
                N = number of tuples in parameter y = numpy.size(x)/2

        Remarks:
            You may set the coefficients before calling the project method
            using projection_func_2d.lcoeff = lcoeff.

            The function calculates for each tuple [yi0,yi1]:
            [xi0,xi1] = scipy.optimize.root( [yi0,yi1] == project([xi0,xi1]) ).x
            Since this is an iterative algorithm, it will be slow.

        '''
        if np.size(lcoeff) > 0: self.lcoeff = lcoeff # set new coefficients if present
        if np.size(luse) > 0: self.luse = luse # set flags
        nx1 = np.size(y)
        assert 0 == nx1%2 and nx1 > 0, 'project expects an even number of coordinates'
        nx = int(nx1 / 2)
        ly = np.reshape(y,(nx,2))
        lx = np.zeros((nx,2), dtype=float) # prepare result arrays
        y0 = self.__lcoeff[0] * self.__luse_term[0] # get shift term for initial guess
        uy0 = self.__luse_term[0]
        d1 = ( self.__lcoeff[1,0] * self.__lcoeff[2,1] \
              - self.__lcoeff[1,1] * self.__lcoeff[2,0] ) \
              * self.__luse_term[1] * self.__luse_term[2]
        m1 = np.array([[1.,0.],[0.,1.]]) # init inverse linear term as identity
        if (d1 != 0.): # get inverse of the linear terms for initial guess
            m1 = np.array([[self.__lcoeff[2,1], -self.__lcoeff[2,0]],[-self.__lcoeff[1,1], self.__lcoeff[1,0]]]) / d1
        self.__luse_term[0] = 1 # activate shift term
        for i in range(0, nx): # for all positions
            self.__lcoeff[0] = y0 - ly[i] # modify shift term
            xguess = np.dot(m1, -self.__lcoeff[0]) # obtain initial guess (extact in first order)
            if (self.max_order > 1):
                # use root finding to get x
                xsol = root(self.__bp_project, xguess, jac=self.__bp_project_deriv_x, tol=tol)
                if xsol.success:
                    lx[i] = xsol.x
                else:
                    lx[i] = xguess
            else: # keep guess in case of a first order projection
                lx[i] = xguess
        # reset lcoeff[0] to initial
        self.__lcoeff[0] = y0
        self.__luse_term[0] = uy0
        return lx

    # --------------------- #
    #                       #
    # least squares methods #
    #                       #
    # --------------------- #

    def __getweights(self, ysigm):
        '''
        Returns an array of weights for an array of error estimates ysigm
        using a mean : individual mixture

        This method is private and should only be called from fit_* methods.

        Parameter:
            ysigm : numpy.array (m,)
                error estimates

        Return:
            weights : numpy.array (m,)
                weights
        '''
        m = np.size(ysigm)
        if (m == 0): return np.array([])
        ysm = np.mean(ysigm)
        if (ysm <= 0.): return np.full(m, 1., dtype=float)
        yss = np.std(ysigm)
        fm = np.min([0.5, yss / ysm]) # mean fraction in weight model
        ys = ysigm * (1. - fm) + fm * ysm
        return np.reciprocal(ys)


    # [x0, lcoeff[1:]] -> prm
    def __getprm_fitx0(self, x0, lcoeff):
        '''
        Return a parameter array for the fitx0 method.
        
        This method is private and should only be called from fit_x0_lcoeff.

        Parameters:
            x0 : numpy.ndarray, shape (2,)
            lcoeff: numpy.ndarray, shape (self.num_terms,2)
        
        Return:
            numpy.ndarray, shape (n,)
        '''
        # all inputs are expected as numpy.ndarray
        luse = self.__luse_term.copy()
        nprm = int(2 * (1 + np.sum(luse[1:])))
        lprm = np.zeros(nprm, dtype=float)
        if np.size(x0) > 1:
            lprm[0:2] = x0[0:2]
        l = 1
        for i in range(1,self.num_terms):
            if luse[i] == 1:
                lprm[2*l:2*l+2] = lcoeff[i,0:2]
                l = l + 1
        return lprm

    # prm -> [x0, lcoeff[1:]]
    def __getcoeff_fitx0(self, lprm):
        '''
        Returs x0 and an lcoeff array for the fitx0 method.

        This method is private and should only be called from fit_x0_lcoeff.
        
        Parameters:
            lprm: numpy.ndarray shape(n,)
            luse: numpy.ndarray shape(self.num_terms,)
        
        Return:
            [ numpy.ndarray shape(2,) , numpy.ndarray shape(self.num_terms,2)
        '''
        luse = self.__luse_term.copy()
        x0 = lprm[0:2]
        lcoeff = np.zeros((self.num_terms, 2))
        l = 1
        for i in range(1,self.num_terms):
            if luse[i] == 1:
                lcoeff[i] = lprm[2*l:2*l+2]
                l = l + 1
        return [x0, lcoeff]

    def __getderiv_fitx0(self, ldx0, ldcoeff):
        '''
        Returns a parameter derivative array for the fitx0 method.

        This method is private and should only be called from fit_x0_lcoeff.
        
        Parameters:
            ldx0 : numpy.ndarray, shape (m,2,2)
            ldcoeff: numpy.ndarray, shape (m,self.num_terms,2,2)
        
        Return:
            numpy.ndarray, shape (m,n/2,2,2), n = total number of parameters
        '''
        luse = self.__luse_term.copy()
        nh = int(1 + np.sum(luse[1:]))
        m = ldx0.shape[0]
        ldprm = np.zeros((m,nh,2,2), dtype=float)
        ldprm[:,0:1,:,:] = ldx0.reshape(m,1,2,2)
        l = 1
        for i in range(1, self.num_terms):
            if luse[i] == 1:
                ldprm[:,l,:,:] = ldcoeff[:,i,:,:]
                l = l + 1
        return ldprm


    # define a function calculating squared differences between measured and
    # modelled peak positions using the 2d projection function
    # prm = [x0, dfobj.lcoeff[1:]]
    def __sdevproj_x0(self,x): # assumes shape(x) = (n,)
        x0, lc = self.__getcoeff_fitx0(x)
        #lc[2:] = x[2:] # copy input
        self.lcoeff = lc # set input
        lx = self.__xdata_fitx0 + x0 # x[0:2] # (m,2) : dxi = xi + x0
        lwf = self.__wdata_fitx0.flatten() # (2*m) : wi
        dpmf = (self.project(lx) - self.__ydata_fitx0).flatten() # (2*m,) : dyi = p(dxi) - yi : [dyi0,dyi1,dy10,dy11,dy20,dy21,...,dyn0,dyn1]
        return np.dot(dpmf, dpmf * lwf) #/ np.sum(lwf) # s = sum_i=1,m[ sum_a=0,1[ dyi**2 * wi ] ]
    # ... and define a Jacobian
    def __dsdevproj_x0(self,x): # assumes shape(x) = (n,)
        x0, lc = self.__getcoeff_fitx0(x)
        #lc = np.zeros(self.num_terms*2)
        #lc[2:] = x[2:] # copy input
        self.lcoeff = lc # set input
        lx = self.__xdata_fitx0 + x0 # x[0:2] # (m,2) : dxi = xi + x0
        lwf = self.__wdata_fitx0.flatten() # (2*m) : wi
        dpm = (self.project(lx) - self.__ydata_fitx0) # (m,2) : dyi = p(dxi) - pi : [[dy00,dy01],[dy10,dy11],[dy20,dy21],...,[dym0,dym1]
        dx0 = self.project_deriv_x(lx) # (m,2,2) : D[dyi,x0] [...,[[ddyi0/dx00, ddyi1/dx00], [ddyi0/dx01 + ddyi1/dx01],...]
        #sdx0 = dx0.shape
        #dx0s = np.reshape(dx0,(sdx0[0],1,sdx0[1],sdx0[2])) # (m,1,2,2)
        dcoeff = self.project_deriv_lcoeff(lx) # (m,n/2,2,2) : D[dyi,aj] = [...,[[ddyi0/daj0,ddyi1/daj0],[ddyi0/daj1,ddyi1/daj1]],...]
        #dp = dcoeff # (m,n/2,2,2)
        #sdp = dp.shape
        #dp[:,0,:,:] = dx0s[:,0,:,:] # (m,n/2,2,2) (replace y shift derivatives by x shift dervatives)
        # ... combine dp and dpm to # (n/2,2) : [...,sum_i[dyi0*ddyi0/daj0 + dyi1*ddyi1/daj0, dyi0*ddyi0/daj1 + dyi1*ddyi1/daj1],...]
        dp = self.__getderiv_fitx0(dx0, dcoeff) # (m,n/2,2,2) : D[dyi,prmj]
        sdp = dp.shape # (m,n/2,2,2)
        # The common dimensions are m = number of samples and 2, number space dimensions.
        dps = np.transpose(dp,(1,3,0,2)).reshape(sdp[1],sdp[3],sdp[0]*sdp[2]) # (n/2,2,2*m) : D[dyi,prmj]
        ds2 = 2. * np.dot(dps, dpm.flatten() * lwf).flatten() # / np.sum(lwf) # (n,) : D[s,prmj] = sum_i=1,m[ 2 * wi * dyi * D[dyi,prmj] ]
        return ds2


    def fit_x0_lcoeff(self, xdata, ydata, ysigm=np.array([]),
                      x0=np.array([0.,0.]), lcoeff0=np.array([]),
                      luse=np.array([])):
        '''
        Fits the projection to a set of data including an x0 shift
        instead of the shift coefficients (order 0) using a minimum
        least-squares approach.

        Parameters:
            xdata : array_like, shape=(m,2)
                object plane position data
            ydata : array_like, shape=(m,2)
                image plane position data
            ysigm : array_like, shape=(m,2), optional
                image plane position error
            x0 : tuple, optional
                x data shift, initializes to [0,0] of not given
            lcoeff : array_like, float, optional
                sequence of projection coefficient interpreted as tuples {[alk0,alk1]}
                initializes to the internal coefficient list of the object if not given
            luse : array_like, int, optional
                polynomial term flags

        Return:
            [prm, cov, msd]
            prm : numpy.ndarray shape(num_terms * 2)
                fitted parameters with (y) shift coefficient at index 0 replaced by the x0 shift
                sorted [x00,x01, ... , alk0, alk1, ...]
            cov : numpy.ndarray shape(num_terms * 2, num_terms * 2)
                covariance matrix, entries are sorted as is prm
            msd : float
                mean square deviation between data and model per data item

        '''
        #
        lcbk = self.lcoeff #  backup coefficients
        lubk = self.luse # backup use flags
        if np.size(lcoeff0) > 0: self.lcoeff = lcoeff0
        if np.size(luse) > 0: self.luse = luse
        self.set_coeff_idx(0, 0., 0.) # no shift
        self.set_use_idx(0, 0) # no shift
        # n = self.num_terms
        # prm0 = np.zeros(2*n, dtype=float)
        # if np.size(x0)>=2: prm0[0:2] = x0[0:2]
        # prm0[2:n] = self.lcoeff.flatten()[2:n] # copy initial parameters (except shift)
        prm0 = self.__getprm_fitx0(x0, self.lcoeff)
        # data
        m2 = min(np.size(xdata),np.size(ydata))
        m = int((m2 - m2%2) / 2) # number of data tuples
        # store data lists as class members
        self.__xdata_fitx0 = np.array(xdata).flatten()[0:2*m].reshape(m,2) # x
        self.__ydata_fitx0 = np.array(ydata).flatten()[0:2*m].reshape(m,2) # y
        self.__wdata_fitx0 = np.full((m,2), 1.) # weights
        # setup weights from y error estimates: wi = 1/sigmai**2
        if np.size(ysigm) >= m2:
            ys = np.array(ysigm).reshape(m,2)
            yst = np.transpose(ys)
            wst = np.array([self.__getweights(yst[0]**2), self.__getweights(yst[1]**2)])
            self.__wdata_fitx0 = np.transpose(wst)
        # minimize weighted square deviations
        nmsol = minimize(self.__sdevproj_x0, prm0, method='BFGS', jac=self.__dsdevproj_x0)
        prmf = nmsol.x
        sqrdev = self.__sdevproj_x0(prmf)
        prmcov = nmsol.hess_inv * sqrdev # np.sum(self.__wdata_fitx0.flatten()) #  rethink this scaling
        # resort result vectors for output
        prml = np.zeros(2 * self.num_terms)
        covl = np.zeros((2 * self.num_terms, 2 * self.num_terms))
        lj = 0
        for j in range(0, self.num_terms): # covl and prml rows
            if (j > 0 and self.luse[j] == 0): continue # skip unused coefficient
            j2 = 2 * j
            lj2 = 2 * lj
            prml[j2:j2+2] = prmf[lj2:lj2+2] # copy prm
            li = 0
            for i in range(0, self.num_terms): # covl columns
                if (i > 0 and self.luse[i] == 0): continue # skip unused coefficient
                i2 = 2 * i
                li2 = 2 * li
                covl[j2:j2+2,i2:i2+2] = prmcov[lj2:lj2+2,li2:li2+2] # copy cov
                li = li + 1
            lj = lj + 1
        # restore backup
        self.lcoeff = lcbk # coefficients
        self.luse = lubk # use flags
        # return fit result
        return [prml, covl, sqrdev]


    # lcoeff[:] -> prm
    def __getlcoeff_lstsq(self, lprm):
        '''
        Sort a parameter array obtained by linear least square parameter fit
        with method fit_lcoeff to internal lcoeff order.

        This method is private and should only be called from fit_lcoeff.
        
        Parameters:
            lprm : numpy.ndarray (n,)
                a list of coefficients as used by the least squares fitting
        
        Return:
            numpy.ndarray, shape (self.num_terms,2)
                a polynomial coefficient list ordered as the internal lcoeff member

        '''
        # all inputs are expected as numpy.ndarray
        luse = self.__luse_term.copy()
        lc = np.zeros((self.num_terms, 2))
        l = 0
        for i in range(0,self.num_terms):
            if luse[i] == 1:
                l2 = 2 * l
                lc[i,0:2] = lprm[l2:l2+2]
                l = l + 1
        return lc

    def __getmat_lstsq(self, xdata, weights=np.array([])):
        '''
        Calculates the derivatives of the polynomial with respect to the
        polynomial coeffcients for given input data and sorts them
        into a matrix for the linear least squares method.

        This method is private and should only be called from fit_lcoeff.

        Parameters:
            xdata : numpy.ndarray shape(m,2)
                list of x data tuples
            weights : numpy.ndarray shape(m,2) optional
                list of weights for each data row
                if weights is not given, or not of the same shape as xdata,
                the errors default to 1

        Return:
            numpy.ndarray shape(2*m, n)
                matrix of the linear equation system to be solved
        '''
        # input is expected as numpy.ndarray
        m = xdata.shape[0]
        lw = np.full(2*m, 1., dtype=float)
        if (xdata.shape == weights.shape):
            lw = weights.flatten()
        ldc = self.project_deriv_lcoeff(xdata) # (m,num_terms,2,2) coefficient derivatives at xdata
        luse = self.luse
        n = int(2 * np.sum(luse)) # number of coefficients
        ldp = np.zeros((2*m,n), dtype=float) # prepare result array
        for i in range (0, m): # loop data rows
            i2 = 2 * i
            l = 0
            for j in range(0, self.num_terms): # loop coefficient columns
                if (luse[j]==0): continue # skip not flagged terms
                l2 = 2 * l
                dcij = ldc[i,j].T # transpose of 2 x 2 coeffcient derivate submatrix
                                  # [[dyi0/daj0, dyi0/daj1],[dyi1/daj0, dyi1/daj1]]
                #print([i,j,l,ldp[i2,l2:l2+2],dcij[0] * lw[i2]])
                ldp[i2,l2:l2+2] = dcij[0] * lw[i2] # copy the weighted  2 x 2 coeffcient derivate submatrix
                #print([i,j,l,ldp[i2+1,l2:l2+2],dcij[1] * lw[i2+1]])
                ldp[i2+1,l2:l2+2] = dcij[1] * lw[i2+1] # copy the weighted  2 x 2 coeffcient derivate submatrix
                l = l + 1
        return ldp

    def fit_lcoeff(self, xdata, ydata, ysigm=np.array([]), luse=np.array([])):
        '''
        Fits the projection to a set of data using a linear least-squares.

        Parameters:
            xdata : array_like, shape=(m,2)
                object plane position data
            ydata : array_like, shape=(m,2)
                image plane position data
            ysigm : array_like, shape=(m,2), optional
                image plane position error
            luse : array_like, int, optional
                polynomial term flags

        Return:
            [prm, cov, msd]
            prm : numpy.ndarray shape(num_terms * 2)
                fitted parameters sorted like the flatted coefficient list lcoeff
                [... ,alk0, alk1, ...]
            cov : numpy.ndarray shape(num_terms * 2, num_terms * 2)
                covariance matrix, entries are sorted as is prm
            msd : float
                mean square deviation between data and model per data item

        '''
        #
        lcbk = self.lcoeff #  backup coefficients
        lubk = self.luse # backup use flags
        if np.size(luse) > 0: self.luse = luse
        # data
        m2 = min(np.size(xdata),np.size(ydata))
        m = int((m2 - m2%2) / 2) # number of data tuples
        # store data lists as private class members
        self.__xdata_fit = np.array(xdata).flatten()[0:2*m].reshape(m,2) # x
        self.__ydata_fit = np.array(ydata).flatten()[0:2*m].reshape(m,2) # y
        self.__wdata_fit = np.full((m,2), 1.) # weights
        # setup weights from y error estimates
        if (np.size(ysigm) >= m2):
            ys = np.array(ysigm).reshape(m,2)
            yst = np.transpose(ys)
            wst = np.array([self.__getweights(yst[0]), self.__getweights(yst[1])])
            self.__wdata_fit = np.transpose(wst)
        # minimize weighted square deviations
        mdesign = self.__getmat_lstsq(self.__xdata_fit, self.__wdata_fit) # (2*m,2*num_terms)
        # print(mdesign)
        vbeta = self.__ydata_fit.reshape(2*m) * self.__wdata_fit.reshape(2*m) # (2*m)
        # print(vbeta)
        #####
        # use a linear least square solver here !
        # but we want one which returns the covariance matrix of the solution as well !
        # this is possible (see numerical recipes chapter 15)
        # but seems not implemented in any of the numpy or scipy codes
        # therefore emilys implements this in emilys.numerics.lstsq.linear_lstsq
        #####
        lsol = llstsq(mdesign, vbeta)
        prmf = lsol.x[0]
        prmcov = lsol.cov
        sqrdev = lsol.chisq[0]
        # resort result vectors for output
        #prml = np.zeros(2 * self.num_terms)
        prml = self.__getlcoeff_lstsq(prmf).flatten()
        covl = np.zeros((2 * self.num_terms, 2 * self.num_terms))
        lj = 0
        for j in range(0, self.num_terms): # covl and prml rows
            if (self.luse[j] == 0): continue # skip unused coefficient
            j2 = 2 * j
            lj2 = 2 * lj
            covl[j2] = self.__getlcoeff_lstsq(prmcov[lj2]).flatten()
            covl[j2+1] = self.__getlcoeff_lstsq(prmcov[lj2+1]).flatten()
            lj = lj + 1
        # restore backup
        self.lcoeff = lcbk # coefficients
        self.luse = lubk # use flags
        # return fit result
        return [prml, covl, sqrdev]


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 03 20:43:00 2019
@author: ju-bar

Functions and classes handling 2D image aberrations

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
from numba import jit # include compilation support
import numpy as np

class aberr_axial_func:
    '''
    
    class aberr_axial_func

    Handles the calculation of axial aberration function values.

    aberration_func objects support arbitrary orders of the aberration
    polynomial. The lowest order is 0 and refers to a shift. Spherical
    aberration C_3,0 is of 3rd order.

    General form of the aberration function:

    Re[ sum_l=1,N sum_k=0,min(l,N-l) C_l+k-1,l-k * w^k * conjugate(w)^l / (l+k) ],
    with 2d coordinates w = w_x + w_y*j and coefficients C_m-1,n = C_m-1,n,x + C_m-1,n,y*j
    expressed by complex numbers.

    '''
    #
    # Coding comment:
    #   Internally, the order of an aberration term is m for an aberration of order m-1.
    #   The rotational symmetry is n, with n = 0 for round aberrations.
    #
    def __init__(self, max_order):
        # maximum polynomial order of the phase terms (max(m) = max_order + 1)
        self.__max_order = max(1, int(max_order)+1) # at least an image shift should be present
        no = self.__max_order
        # number of aberration terms
        self.__num_terms = (int)((4 * no - no%2 + no**2) / 4)
        nt = self.__num_terms
        # number of indices for partial term orders, must support 0 and max(m) + max(n) = 2 * max(m)
        self.__lcoeff = np.zeros(nt, dtype=complex) # preset to 0
        # aberration term flags
        self.__luse_term = np.full(nt, 1, dtype=int) # preset to 1
        # aberration coefficient list (internal sequence -> (m,n) )
        self.__lterm_of_idx = np.zeros((nt, 2), dtype=int) # preset to 0
        # aberration index list ( (m,n) -> internal sequence)
        self.__lidx_of_term = np.zeros((no + 1, no + 1), dtype=int) # preset to 0
        # set index list
        l = 0
        for m in range(1, no + 1):
            for n in range(0, m + 1):
                if 0 == (m+n)%2: # valid orders have even sums of m and n
                    self.__lterm_of_idx[l,0] = m
                    self.__lterm_of_idx[l,1] = n
                    self.__lidx_of_term[m,n] = l
                    l = l + 1

    @property
    def max_order(self):
        '''
        Maximum order of the aberation function polynomial.
        '''
        return self.__max_order

    @property
    def num_terms(self):
        '''
        Number of aberration terms.
        '''
        return self.__num_terms

    def idx_of_term(self, m, n):
        '''
        List index of the term with polynomial order (m+1)
        and rotational symmetry n.
        '''
        im = m+1
        if 0 == (im+n)%2 and im <= self.max_order and n <= im:
            return self.__lidx_of_term[im, n]
        return -1 # not a listed term

    def term_of_idx(self, idx):
        '''
        Term order m and rotational symmetry for list index idx
        returned as tuple (m,n).
        '''
        if idx >= 0 and idx < self.num_terms:
            return self.__lterm_of_idx[idx] - [1,0]
        return [0,0] # not a listed term

    @property
    def lcoeff(self):
        '''
        Returns a copy of the internal coefficient list.
        '''
        return self.__lcoeff.copy()

    @lcoeff.setter
    def lcoeff(self, lc):
        '''
        Set internal coefficient list.

        Parameters:
            lcoeff : array_like, complex, shape = (self.num_terms,) or shorter
                aberration coefficients

        Return:
            integer, number of coefficients set
        '''
        lcoeff = np.atleast_1d(lc)
        assert lcoeff.dtype == complex, 'lcoeff has to be of complex numbers!'
        ncf = min(np.size(lcoeff), self.num_terms)
        if ncf > 0:
            self.__lcoeff[0:ncf] = lcoeff[0:ncf]
        #return ncf


    @property
    def luse_term(self):
        '''
        Returns a copy of the internal list of term usage flags.
        '''
        return self.__luse_term.copy()

    @luse_term.setter
    def luse_term(self, luse):
        '''
        Set internal coefficient list.

        Parameters:
            luseterm : array_like, int, shape = (self.num_terms,) or shorter
                aberration term flags

        Return:
            integer, number of coefficient usage flags set
        '''
        luse2 = np.atleast_1d(luse)
        assert luse2.dtype == int, 'luse has to be of int numbers!'
        nuse = min(np.size(luse2), self.num_terms)
        if nuse > 0:
            self.__luse_term[0:nuse] = luse2[0:nuse]
        #return nuse
    
    
    def chi(self, x, lcoeff = None, luse = None):
        '''
        Calculates the aberration function polynomial chi.
        If non-empty lists are provided by parameters 'lcoeff' and 'luse',
        respective internal lists of coefficients and use flags will be set
        from input. Empty lists are default.
        
        You may preset the coefficients by self.lcoeff = [ complex values ]
        and the use flags by self.luse_term = [ integer flags ].

        Parameters:
            x : complex
                coordinate
            lcoeff : array_like, complex, shape = (self.num_coeff(),) or shorter
                aberration coefficients
            luse : array_like, int, shape = (self.num_terms(),) or shorter
                aberration term flags

        Return:
            float

        '''
        reschi = 0.
        nol = self.__max_order + 1
        if lcoeff != None: self.lcoeff = lcoeff
        if luse != None: self.luse_term = luse
        px = 1.
        xfn = np.full(nol, 1., dtype=complex)
        for i in range(1, nol):
            px *= x
            xfn[i] = px
        xfc = xfn.conjugate()
        for k in range(0, self.num_terms): # loop over aberration terms
            if 0 == self.__luse_term[k]:
                continue # skip term
            # adding term (m,n)
            m1, n = self.term_of_idx(k)
            m = m1 + 1
            j = int((m + n) / 2)
            l = int((m - n) / 2)
            term = np.real( self.__lcoeff[k] * xfn[l] * xfc[j] ) / m
            reschi = reschi + term # add term to result
        return reschi

    #@jit
    def chi_list(self, lx, lcoeff = None, luse = None):
        '''
        Calculates the aberration function polynomial chi for a list
        of coordinates lx.
        If non-empty lists are provided by parameters 'lcoeff' and 'luse',
        respective internal lists of coefficients and use flags will be set
        from input. Empty lists are default.
        
        You may preset the coefficients by self.lcoeff = [ complex values ]
        and the use flags by self.luse_term = [ integer flags ].

        Parameters:
            lx : array_like, complex
                list of coordinate
            lcoeff : array_like, complex, shape = (self.num_coeff(),) or shorter
                aberration coefficients
            luse : array_like, int, shape = (self.num_terms(),) or shorter
                aberration term flags

        Return:
            np.array, size = size(lx), float

        '''
        nx = np.size(lx)
        if nx == 0:
            return np.array([], dtype=float)
        lc = np.zeros(nx, dtype=float)
        rchi = 0.
        nol = self.__max_order + 1
        if lcoeff != None: self.lcoeff = lcoeff
        if luse != None: self.luse_term = luse
        xfn = np.full(nol, 1., dtype=complex)
        #
        # loop over coordinates
        for i in range(0, nx):
            x = lx[i]
            px = 1.
            for j in range(1, nol):
                px *= x
                xfn[j] = px
            xfc = xfn.conjugate()
            for k in range(0, self.num_terms): # loop over aberration terms
                if 0 == self.__luse_term[k]:
                    continue # skip term
                # adding term (m,n)
                m1, n = self.term_of_idx(k)
                m = m1 + 1
                j = int((m + n) / 2)
                l = int((m - n) / 2)
                term = np.real( self.__lcoeff[k] * xfn[l] * xfc[j] ) / m
                rchi = rchi + term # add term to result
            lc[i] = rchi
        return lc

    def grad_chi(self, x, lcoeff = None, luse = None):
        '''
        Calculates the gradient of aberration function polynomial chi
        in complex number notation.
        If non-empty lists are provided by parameters 'lcoeff' and 'luse',
        respective internal lists of coefficients and use flags will be set
        from input. Empty lists are default.

        You may preset the coefficients by self.lcoeff = [ complex values ]
        and the use flags by self.luse_term = [ integer flags ].
        
        Parameters:
            x : complex
                coordinate
            lcoeff : array_like, complex, shape = (self.num_coeff(),) or shorter
                aberration coefficients
            luse : array_like, int, shape = (self.num_terms(),) or shorter
                aberration term flags

        Return:
            complex

        '''
        gradchi = 0. + 0.j
        nol = self.__max_order + 2
        if lcoeff != None: self.lcoeff = lcoeff
        if luse != None: self.luse_term = luse
        px = 1.
        # initialize complex powers [0., 1., x, x^2, x^3, ...]
        xfn = np.zeros(nol, dtype=complex)
        xfn[1] = 1.
        for i in range(2, nol):
            px *= x
            xfn[i] = px
        xfc = xfn.conjugate() # conjugates of the power list
        for k in range(0, self.num_terms): # loop over aberration terms
            if 0 == self.__luse_term[k]:
                continue # skip term
            # adding term (m,n)
            m1, n = self.term_of_idx(k)
            m = m1 + 1
            m2 = 2*m
            j = int((m + n) / 2)
            l = int((m - n) / 2)
            # remember: index 0 -> 0, 1 -> x^0, 2 -> x^1, 3 -> x^2 ...
            #   therefore l -> l+1, j -> j+1 compared to the series in method chi(...)
            term = self.__lcoeff[k] * xfn[l+1] * xfc[j] * j \
                   + np.conjugate(self.__lcoeff[k]) * xfc[l] * xfn[j+1] * l
            gradchi = gradchi + term / m2 # add term to result
        return gradchi

    #@jit
    def grad_chi_list(self, lx, lcoeff = None, luse = None):
        '''
        Calculates the gradient of aberration function polynomial chi
        in complex number notation for a list of coordinates lx.
        If non-empty lists are provided by parameters 'lcoeff' and 'luse',
        respective internal lists of coefficients and use flags will be set
        from input. Empty lists are default.

        You may preset the coefficients by self.lcoeff = [ complex values ]
        and the use flags by self.luse_term = [ integer flags ].
        
        Parameters:
            lx : array_like, complex
                coordinates
            lcoeff : array_like, complex, shape = (self.num_coeff(),) or shorter
                aberration coefficients
            luse : array_like, int, shape = (self.num_terms(),) or shorter
                aberration term flags

        Return:
            np.array, size=size(lx), complex

        '''
        nx = np.size(lx)
        if nx == 0:
            return np.array([], dtype=complex)
        lg = np.zeros(nx, dtype=complex)
        gradchi = 0. + 0.j
        nol = self.__max_order + 2
        if lcoeff != None: self.lcoeff = lcoeff
        if luse != None: self.luse_term = luse
        # initialize complex powers [0., 1., x, x^2, x^3, ...]
        xfn = np.zeros(nol, dtype=complex)
        xfn[1] = 1.
        #
        # loop over list
        for i in range(0, nx):
            x = lx[i] # get current coordinate
            # prepare powers of x
            px = 1.
            for j in range(2, nol):
                px *= x
                xfn[j] = px
            xfc = xfn.conjugate() # conjugates of the power list
            gradchi = 0. + 0.j # reset current result
            for k in range(0, self.num_terms): # loop over aberration terms
                if 0 == self.__luse_term[k]:
                    continue # skip term
                # adding term (m,n)
                m1, n = self.term_of_idx(k)
                m = m1 + 1
                m2 = 2*m
                j = int((m + n) / 2)
                l = int((m - n) / 2)
                # remember: index 0 -> 0, 1 -> x^0, 2 -> x^1, 3 -> x^2 ...
                #   therefore l -> l+1, j -> j+1 compared to the series in method chi(...)
                term = self.__lcoeff[k] * xfn[l+1] * xfc[j] * j \
                       + np.conjugate(self.__lcoeff[k]) * xfc[l] * xfn[j+1] * l
                gradchi = gradchi + term / m2 # add term to result
            lg[i] = gradchi
        return lg



# -*- coding: utf-8 -*-
"""
Created on Sun Jul 1 12:30:00 2019
@author: ju-bar

Monte-Carlo tools

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
# %%
from numba import jit # include compilation support
import numpy as np # include numeric functions
import emilys.image.imagedata as img # include image data routines
# %%
@jit
def mc_image_pos_maximize(image, lpoints, rstep=1., etmp=0., itmax=100):
    '''

    Maximizes the intensity given by image for points in lpoints by
    moving the points in a Monte-Carlo approach.

    Returns:
        >0 = number of iterations performed.
        <0 = error code
        lpoints = points close to maximum positions

    '''
    dimg = image.shape
    dpts = lpoints.shape
    if len(dimg) != 2:
        return -10
    if len(dpts) != 2:
        return -11
    if dpts[1] != 2:
        return -12
    npts = dpts[0]
    e0 = img.image_pos_sum(image, lpoints)
    et = e0
    s = np.abs(rstep)
    it = 0
    repeat = 1
    while repeat == 1:
        ep = img.image_pos_sum(image, lpoints) # previous samples intensity
        et = ep # sampled intensity of this run
        for l in range(0,npts):
            epl = img.image_at(image, lpoints[l])
            dpos = np.random.normal(size=2) * s
            tpos = lpoints[l] + dpos
            ept = img.image_at(image, tpos)
            epd = ept - epl
            if epd > 0.: # improve
                lpoints[l] = tpos
                et = et + epd
            elif etmp > 0: # edp <= 0 maybe try
                win = epd / etmp
                pacc = 0.
                if win > -200.:
                    pacc = min(1., np.exp(win)) # finite probability to accept 0 ... 1
                if pacc > np.random.random_sample(): # accept the worse try
                    lpoints[l] = tpos
                    et = et + epd
        de = et - ep
        it = it + 1
        if it >= itmax: repeat = 0
        if de < etmp: repeat = 0 # converged
    return it


import setuptools

setuptools.setup(
    name="emilys",
    version='0.1.0',
    url='https://github.com/ju-bar/emilys',
    packages=setuptools.find_packages(),
    author='Juri Barthel',
    author_email='juribarthel@gmail.com',
    description='Python scripts for Electron Microscopy Image anaLYsis'
)


# -*- coding: utf-8 -*-
"""
Created on Fri Jul 12 15:57:00 2019
@author: ju-bar

Peak fitting

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
import numpy as np
from scipy.optimize import curve_fit
import emilys.functions.peaks as pks
# %%
def get_data_in_circroi(image, pos, rad, wrap=True):
    '''
    Extracts data points in a circular region of interest from an image.

    Parameters:
        image : numpy.ndarray
            image data
        pos : array (x,y)
            center position for the fit region of interest
        rad : float
            radius of the fit region of interest
            must be larger than 2
        wrap : bool
            flag for using periodic wrap around when crossing
            input image bounds

    Return:
        [number of image points n, positions, image values]
        [int, numpy.ndarray (n,2), numpy.ndarray (n)]
    '''
    ndim = np.array(image.shape)
    assert ndim.size == 2, 'expecting a 2d array as parameter 1'
    assert np.size(pos) == 2, 'expecting a tuple as parameter 2'
    assert np.abs(rad) > 2, 'expecting rad > 2 (parameter 3)'
    icent = np.array(pos).astype(int)
    irad = np.ceil(np.abs(rad)) + 1
    ir0 = icent - np.array([irad,irad], dtype=int)
    ir1 = icent + np.array([irad,irad], dtype=int)
    ndat = (ir1[0] - ir0[0] + 1) * (ir1[1] - ir0[1] + 1)
    datx = np.zeros((ndat,2), dtype=float)
    daty = np.zeros(ndat, dtype=float)
    nfit = 0
    for k in range(ir0[1], ir1[1]+1):
        ik = k
        if wrap:
            ik = int( k%ndim[0] )
        else:
            if (k < 0 or k >= ndim[0]):
                continue # skip this row
        dy = k - pos[1]
        dy2 = dy**2
        for h in range(ir0[0], ir1[0]+1):
            ih = h
            if wrap:
                ih = int( h%ndim[1] )
            else:
                if (h < 0 or h >= ndim[1]):
                    continue # skip this columns
            dx = h - pos[0]
            d = np.sqrt(dy2 + dx**2)
            if (d <= np.abs(rad) and nfit < ndat):
                datx[nfit] = np.array([h,k])
                daty[nfit] = image[ik,ih]
                nfit = nfit + 1
    if (nfit > 0):
        return [nfit, datx[0:nfit], daty[0:nfit]]
    return [0, [], []]
# %%
def get_values_in_circroi(image, pos, rad, wrap=True):
    '''
    Extracts values in a circular region of interest from an image.

    Parameters:
        image : numpy.ndarray
            image data
        pos : array (x,y)
            center position for the fit region of interest
        rad : float
            radius of the fit region of interest
            must be larger than 2
        wrap : bool
            flag for using periodic wrap around when crossing
            input image bounds

    Return:
        [number of image points n, image values]
        [int, numpy.ndarray (n)]
    '''
    ndim = np.array(image.shape)
    assert ndim.size == 2, 'expecting a 2d array as parameter 1'
    assert np.size(pos) == 2, 'expecting a tuple as parameter 2'
    assert np.abs(rad) > 2, 'expecting rad > 2 (parameter 3)'
    icent = np.array(pos).astype(int)
    irad = np.ceil(np.abs(rad)) + 1
    ir0 = icent - np.array([irad,irad], dtype=int)
    ir1 = icent + np.array([irad,irad], dtype=int)
    ndat = (ir1[0] - ir0[0] + 1) * (ir1[1] - ir0[1] + 1)
    daty = np.zeros(ndat, dtype=float)
    nfit = 0
    for k in range(ir0[1], ir1[1]+1):
        ik = k
        if wrap:
            ik = int( k%ndim[0] )
        else:
            if (k < 0 or k >= ndim[0]):
                continue # skip this row
        dy = k - pos[1]
        dy2 = dy**2
        for h in range(ir0[0], ir1[0]+1):
            ih = h
            if wrap:
                ih = int( h%ndim[1] )
            else:
                if (h < 0 or h >= ndim[1]):
                    continue # skip this columns
            dx = h - pos[0]
            d = np.sqrt(dy2 + dx**2)
            if (d <= np.abs(rad) and nfit < ndat):
                daty[nfit] = image[ik,ih]
                nfit = nfit + 1
    if (nfit > 0):
        return [nfit, daty[0:nfit]]
    return [0, []]
# %%
def fit_local_gauss_2d(image, pos, rad, wrap=True, imagesigma=None):
    '''
    Fits a gaussian peak to a circular local image area.

    Parameters:
        image : numpy.ndarray
            image data
        pos : array (x,y)
            center position for the fit region of interest
        rad : float
            radius of the fit region of interest
            must be larger than 2
        wrap : bool
            flag for using periodic wrap around when crossing
            input image bounds

    Return:
        [best fitting parameters, parameter error estimates]
        fit parameters: [x0, y0, a, bxx, bxy, byy, c], 
        see emilys.functions.peaks.gauss_2d
    '''
    nprm = 7 # number of peak parameters
    ndim = np.array(image.shape)
    assert ndim.size == 2, 'expecting a 2d array as parameter 1'
    assert np.size(pos) == 2, 'expecting a tuple as parameter 2'
    assert np.abs(rad) > 2, 'expecting rad > 2 (parameter 3)'
    (nfit, xfit, yfit) = get_data_in_circroi(image, pos, rad, wrap)
    assert nfit > 2 * nprm, 'insufficient number of image points '\
        '({:d} of {:d}) in region of interest'.format(nfit, 2*nprm)
    ysig = None
    if (imagesigma != None): # also extract image sigma values
        (nfit2, ysig) = get_values_in_circroi(imagesigma, pos, rad, wrap)
        assert nfit == nfit2, 'internal data size conflict with sigma data'
    prm0 = [ pos[0], pos[1], np.max(yfit) - np.min(yfit), 
             1./np.abs(rad), 0., 1./np.abs(rad), np.min(yfit)] # initial parameter set
    solprm, solcov = curve_fit( pks.gauss_2d, xfit.T, yfit, prm0,
                                jac=pks.gauss_2d_jac, sigma=ysig) # call scipy.optimize.curve_fit
    solerr = np.sqrt(np.diag(solcov))
    return [solprm, solerr]